# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re
from blockdiag.utils import Box, Size, XY
from blockdiag.utils.compat import u, string_types


def splitlabel(string):
    """Split text to lines as generator.
       Every line will be stripped.
       If text includes characters "\n", treat as line separator.
    """
    string = re.sub('^\s*', '', string)
    string = re.sub('\s*$', '', string)
    string = re.sub('\xa5', '\\\\', string)
    string = re.sub('(\\\\){2}', '\x00', string)
    string = re.sub('\\\\n', '\n', string)
    for line in string.splitlines():
        yield re.sub('\x00', '\\\\', line).strip()


def splittext(metrics, text, bound, measure='width'):
    folded = []
    if text == '':
        folded.append(u(' '))

    for i in range(len(text), 0, -1):
        textsize = metrics.textsize(text[0:i])

        if getattr(textsize, measure) <= bound:
            folded.append(text[0:i])
            if text[i:]:
                folded += splittext(metrics, text[i:], bound, measure)
            break

    return folded


def truncate_text(metrics, text, bound, measure='width'):
    for i in range(len(text), 0, -1):
        textsize = metrics.textsize(text[0:i] + ' ...')

        if getattr(textsize, measure) <= bound:
            return text[0:i] + ' ...'

    return text


def get(*args, **kwargs):
    if kwargs.get('orientation') == 'vertical':
        return VerticalTextFolder(*args, **kwargs)
    else:
        return HorizontalTextFolder(*args, **kwargs)


class VerticalTextFolder(object):
    def __init__(self, drawer, box, string, font, **kwargs):
        self.drawer = drawer
        self.box = box
        self.string = string
        self.font = font
        self.scale = 1
        self.halign = kwargs.get('halign', 'center')
        self.valign = kwargs.get('valign', 'center')
        self.padding = kwargs.get('padding', 8)
        self.line_spacing = kwargs.get('line_spacing', 2)

        if kwargs.get('adjustBaseline'):
            self.adjustBaseline = True
        else:
            self.adjustBaseline = False

        self._result = self._lines()

    def textsize(self, text, scaled=False):
        if isinstance(text, string_types):
            size = [self.drawer.textlinesize(c, self.font) for c in text]
            width = max(s.width for s in size)
            height = (sum(s.height for s in size) +
                      self.line_spacing * (len(text) - 1))

            textsize = Size(width, height)
        else:
            if text:
                size = [self.textsize(s) for s in text]
                height = max(s.height for s in size)
                width = (sum(s.width for s in size) +
                         self.line_spacing * (len(text) - 1))

                textsize = Size(width, height)
            else:
                textsize = Size(0, 0)

        if scaled:
            textsize = Size(textsize.width * self.scale,
                            textsize.height * self.scale)

        return textsize

    @property
    def lines(self):
        textsize = self.textsize(self._result, scaled=True)

        dx, _ = self.box.get_padding_for(textsize, halign=self.halign,
                                         padding=self.padding)

        width = self.box.width - dx + self.line_spacing
        base_xy = XY(self.box.x1, self.box.y1)
        for string in self._result:
            textsize = self.textsize(string, scaled=True)
            _, dy = self.box.get_padding_for(textsize, valign=self.valign,
                                             padding=self.line_spacing)

            height = dy
            width -= textsize.width + self.line_spacing
            for char in string:
                charsize = self.textsize(char, scaled=True)

                if self.adjustBaseline:
                    draw_xy = base_xy.shift(width, height + charsize.height)
                else:
                    draw_xy = base_xy.shift(width, height)

                yield char, draw_xy

                height += charsize.height + self.line_spacing

    @property
    def outlinebox(self):
        corners = []
        for string, xy in self.lines:
            textsize = self.textsize(string)
            width = textsize[0] * self.scale
            height = textsize[1] * self.scale

            if self.adjustBaseline:
                xy = XY(xy.x, xy.y - textsize[1])

            corners.append(xy)
            corners.append(XY(xy.x + width, xy.y + height))

        if corners:
            box = Box(min(p.x for p in corners) - self.padding,
                      min(p.y for p in corners) - self.line_spacing,
                      max(p.x for p in corners) + self.padding,
                      max(p.y for p in corners) + self.line_spacing)
        else:
            box = Box(self.box[0], self.box[1], self.box[0], self.box[1])

        return box

    def _lines(self):
        lines = []
        measure = 'height'
        maxwidth, maxheight = self.box.size

        width = 0
        finished = False
        for line in splitlabel(self.string):
            for folded in splittext(self, line, maxheight, measure):
                textsize = self.textsize(folded)

                if width + textsize.width + self.line_spacing < maxwidth:
                    lines.append(folded)
                    width += textsize.width + self.line_spacing
                elif len(lines) > 0:
                    lines[-1] = truncate_text(self, lines[-1],
                                              maxheight, measure)
                    finished = True
                    break

            if finished:
                break

        return lines


class HorizontalTextFolder(object):
    def __init__(self, drawer, box, string, font, **kwargs):
        self.drawer = drawer
        self.box = box
        self.string = string
        self.font = font
        self.scale = 1
        self.halign = kwargs.get('halign', 'center')
        self.valign = kwargs.get('valign', 'center')
        self.padding = kwargs.get('padding', 8)
        self.line_spacing = kwargs.get('line_spacing', 2)

        if kwargs.get('adjustBaseline'):
            self.adjustBaseline = True
        else:
            self.adjustBaseline = False

        self._result = self._lines()

    def textsize(self, text, scaled=False):
        if isinstance(text, string_types):
            textsize = self.drawer.textlinesize(text, self.font)
        else:
            if text:
                size = [self.textsize(s) for s in text]
                width = max(s.width for s in size)
                height = (sum(s.height for s in size) +
                          self.line_spacing * (len(text) - 1))

                textsize = Size(width, height)
            else:
                textsize = Size(0, 0)

        if scaled:
            textsize = Size(textsize.width * self.scale,
                            textsize.height * self.scale)

        return textsize

    @property
    def lines(self):
        textsize = self.textsize(self._result, scaled=True)

        _, dy = self.box.get_padding_for(textsize, valign=self.valign,
                                         padding=self.line_spacing)

        height = dy
        base_xy = XY(self.box.x1, self.box.y1)
        for string in self._result:
            textsize = self.textsize(string, scaled=True)
            dx, _ = self.box.get_padding_for(textsize, halign=self.halign,
                                             padding=self.padding)

            if self.adjustBaseline:
                draw_xy = base_xy.shift(dx, height + textsize.height)
            else:
                draw_xy = base_xy.shift(dx, height)

            yield string, draw_xy

            height += textsize.height + self.line_spacing

    @property
    def outlinebox(self):
        corners = []
        for string, xy in self.lines:
            textsize = self.textsize(string)
            width = textsize[0] * self.scale
            height = textsize[1] * self.scale

            if self.adjustBaseline:
                xy = XY(xy.x, xy.y - textsize[1])

            corners.append(xy)
            corners.append(XY(xy.x + width, xy.y + height))

        if corners:
            box = Box(min(p.x for p in corners) - self.padding,
                      min(p.y for p in corners) - self.line_spacing,
                      max(p.x for p in corners) + self.padding,
                      max(p.y for p in corners) + self.line_spacing)
        else:
            box = Box(self.box[0], self.box[1], self.box[0], self.box[1])

        return box

    def _lines(self):
        lines = []
        measure = 'width'
        maxwidth, maxheight = self.box.size

        height = 0
        finished = False
        for line in splitlabel(self.string):
            for folded in splittext(self, line, maxwidth, measure):
                textsize = self.textsize(folded)

                if height + textsize.height + self.line_spacing < maxheight:
                    lines.append(folded)
                    height += textsize.height + self.line_spacing
                else:
                    if len(lines) > 0:
                        lines[-1] = truncate_text(self, lines[-1],
                                                  maxwidth, measure)

                    finished = True
                    break

            if finished:
                break

        return lines



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import re
import sys
import traceback
from optparse import OptionParser, SUPPRESS_HELP
from blockdiag import imagedraw
from blockdiag import plugins
from blockdiag.utils import images
from blockdiag.utils.compat import codecs
from blockdiag.utils.config import ConfigParser
from blockdiag.utils.fontmap import parse_fontpath, FontMap
from blockdiag.utils.logging import warning, error


class Application(object):
    module = None
    options = None

    def __init__(self):
        self.cleanup_handlers = []

    def __enter__(self):
        self.setup()
        return self

    def __exit__(self, *args):
        self.cleanup()

    def register_cleanup_handler(self, handler):
        self.cleanup_handlers.append(handler)

    def run(self, args):
        try:
            self.parse_options(args)
            self.create_fontmap()
            self.setup()

            parsed = self.parse_diagram()
            return self.build_diagram(parsed)
        except SystemExit as e:
            return e
        except UnicodeEncodeError as e:
            error("UnicodeEncodeError caught (check your font settings)")
            return -1
        except Exception as e:
            if self.options and self.options.debug:
                traceback.print_exc()
            else:
                error("%s" % e)
            return -1
        finally:
            self.cleanup()

    def parse_options(self, args):
        self.options = Options(self.module).parse(args)

    def create_fontmap(self):
        self.fontmap = create_fontmap(self.options)

    def setup(self):
        images.setup(self)
        plugins.setup(self)

    def parse_diagram(self):
        if self.options.input == '-':
            stream = codecs.getreader('utf-8-sig')(sys.stdin)
            self.code = stream.read()
        else:
            fp = codecs.open(self.options.input, 'r', 'utf-8-sig')
            self.code = fp.read()

        return self.module.parser.parse_string(self.code)

    def build_diagram(self, tree):
        ScreenNodeBuilder = self.module.builder.ScreenNodeBuilder
        try:
            diagram = ScreenNodeBuilder.build(tree, self.options)
        except:
            diagram = ScreenNodeBuilder.build(tree)  # old interface

        DiagramDraw = self.module.drawer.DiagramDraw
        drawer = DiagramDraw(self.options.type, diagram,
                             self.options.output, fontmap=self.fontmap,
                             code=self.code, antialias=self.options.antialias,
                             nodoctype=self.options.nodoctype,
                             transparency=self.options.transparency)
        drawer.draw()

        if self.options.size:
            drawer.save(size=self.options.size)
        else:
            drawer.save()

        return 0

    def cleanup(self):
        for handler in self.cleanup_handlers[:]:
            try:
                handler()
            except Exception as exc:
                error("%s" % exc)
            finally:
                self.cleanup_handlers.remove(handler)


class Options(object):
    def __init__(self, module):
        self.module = module
        self.build_parser()

    def parse(self, args):
        self.options, self.args = self.parser.parse_args(args)
        self.validate()
        self.read_configfile()

        return self.options

    def build_parser(self):
        version = "%%prog %s" % self.module.__version__
        usage = "usage: %prog [options] infile"
        self.parser = p = OptionParser(usage=usage, version=version)
        p.add_option('-a', '--antialias', action='store_true',
                     help='Pass diagram image to anti-alias filter')
        p.add_option('-c', '--config',
                     help='read configurations from FILE', metavar='FILE')
        p.add_option('--debug', action='store_true',
                     help='Enable debug mode')
        p.add_option('-o', dest='output',
                     help='write diagram to FILE', metavar='FILE')
        p.add_option('-f', '--font', default=[], action='append',
                     help='use FONT to draw diagram', metavar='FONT')
        p.add_option('--fontmap',
                     help='use FONTMAP file to draw diagram', metavar='FONT')
        p.add_option('--ignore-pil', dest='ignore_pil',
                     default=False, action='store_true', help=SUPPRESS_HELP)
        p.add_option('--no-transparency', dest='transparency',
                     default=True, action='store_false',
                     help='do not make transparent background of diagram ' +
                          '(PNG only)')
        p.add_option('--size',
                     help='Size of diagram (ex. 320x240)')
        p.add_option('-T', dest='type', default='PNG',
                     help='Output diagram as TYPE format')
        p.add_option('--nodoctype', action='store_true',
                     help='Do not output doctype definition tags (SVG only)')

        return p

    def validate(self):
        if len(self.args) == 0:
            self.parser.print_help()
            sys.exit(0)

        self.options.input = self.args.pop(0)
        if self.options.output:
            pass
        elif self.options.output == '-':
            self.options.output = 'output.' + self.options.type.lower()
        else:
            basename = os.path.splitext(self.options.input)[0]
            ext = '.%s' % self.options.type.lower()
            self.options.output = basename + ext

        self.options.type = self.options.type.upper()
        try:
            imagedraw.create(self.options.type, None, debug=self.options.debug)
        except:
            msg = "unknown format: %s" % self.options.type
            raise RuntimeError(msg)

        if self.options.size:
            matched = re.match('^(\d+)x(\d+)$', self.options.size)
            if matched:
                self.options.size = [int(n) for n in matched.groups()]
            else:
                msg = "--size option must be formatted as WIDTHxHEIGHT."
                raise RuntimeError(msg)

        if self.options.type == 'PDF':
            try:
                import reportlab.pdfgen.canvas
                reportlab.pdfgen.canvas
            except ImportError:
                msg = "could not output PDF format; Install reportlab."
                raise RuntimeError(msg)

        if self.options.ignore_pil:
            warning("--ignore-pil option is deprecated "
                    "(detect automatically).")

        if self.options.nodoctype and self.options.type != 'SVG':
            msg = "--nodoctype option work in SVG images."
            raise RuntimeError(msg)

        if self.options.transparency is False and self.options.type != 'PNG':
            msg = "--no-transparency option work in PNG images."
            raise RuntimeError(msg)

        if self.options.config and not os.path.isfile(self.options.config):
            msg = "config file is not found: %s" % self.options.config
            raise RuntimeError(msg)

        if self.options.fontmap and not os.path.isfile(self.options.fontmap):
            msg = "fontmap file is not found: %s" % self.options.fontmap
            raise RuntimeError(msg)

    def read_configfile(self):
        if self.options.config:
            configpath = self.options.config
        elif os.environ.get('HOME'):
            configpath = '%s/.blockdiagrc' % os.environ.get('HOME')
        elif os.environ.get('USERPROFILE'):
            configpath = '%s/.blockdiagrc' % os.environ.get('USERPROFILE')
        else:
            configpath = ''

        appname = self.module.__name__
        if os.path.isfile(configpath):
            config = ConfigParser()
            config.read(configpath)

            if config.has_option(appname, 'fontpath'):
                fontpath = config.get(appname, 'fontpath')
                self.options.font.append(fontpath)

            if config.has_option(appname, 'fontmap'):
                if self.options.fontmap is None:
                    self.options.fontmap = config.get(appname, 'fontmap')

            if config.has_option(appname, 'antialias'):
                antialias = config.get(appname, 'antialias')
                if antialias.lower() == 'true':
                    self.options.antialias = True

            if self.options.fontmap is None:
                self.options.fontmap = configpath


def detectfont(options):
    import glob
    fontdirs = [
        '/usr/share/fonts',
        '/Library/Fonts',
        '/System/Library/Fonts',
        'c:/windows/fonts',
        '/usr/local/share/font-*',
    ]
    fontfiles = [
        'ipagp.ttf',
        'ipagp.otf',
        'VL-PGothic-Regular.ttf',
        'Hiragino Sans GB W3.otf',
        'AppleGothic.ttf',
        'msgothic.ttf',
        'msgoth04.ttf',
        'msgothic.ttc',
    ]

    fontpath = None
    if options.font:
        for path in options.font:
            _path, _ = parse_fontpath(path)
            if os.path.isfile(_path):
                fontpath = path
                break
        else:
            msg = 'fontfile is not found: %s' % options.font
            raise RuntimeError(msg)

    if fontpath is None:
        globber = (glob.glob(d) for d in fontdirs)
        for fontdir in sum(globber, []):
            for root, _, files in os.walk(fontdir):
                for font in fontfiles:
                    if font in files:
                        fontpath = os.path.join(root, font)
                        break

    return fontpath


def create_fontmap(options):
    fontmap = FontMap(options.fontmap)
    if fontmap.find().path is None or options.font:
        fontpath = detectfont(options)
        fontmap.set_default_font(fontpath)

    return fontmap



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from blockdiag import plugins


class NodeAttributes(plugins.NodeHandler):
    def __init__(self, diagram, **kwargs):
        super(NodeAttributes, self).__init__(diagram, **kwargs)

        node_klass = diagram._DiagramNode
        for name, label in kwargs.items():
            if label is None:
                label = name
            if name not in node_klass.desctable:
                node_klass.desctable.insert(-1, name)

            node_klass.attrname[name] = label
            if not hasattr(node_klass, name):
                setattr(node_klass, name, None)


def setup(self, diagram, **kwargs):
    plugins.install_node_handler(NodeAttributes(diagram, **kwargs))



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from blockdiag.tests.utils import BuilderTestCase, capture_stderr


class TestBuilderEdge(BuilderTestCase):
    def test_diagram_attributes(self):
        diagram = self.build('diagram_attributes.diag')
        self.assertEqual(2, len(diagram.nodes))
        self.assertEqual(1, len(diagram.edges))

    def test_single_edge_diagram(self):
        diagram = self.build('single_edge.diag')
        self.assertEqual(2, len(diagram.nodes))
        self.assertEqual(1, len(diagram.edges))

        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0)})
        self.assertNodeLabel(diagram, {'A': 'A', 'B': 'B'})

    def test_two_edges_diagram(self):
        diagram = self.build('two_edges.diag')
        self.assertEqual(3, len(diagram.nodes))
        self.assertEqual(2, len(diagram.edges))

        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0), 'C': (2, 0)})

    def test_edge_shape(self):
        diagram = self.build('edge_shape.diag')
        self.assertEdgeDir(diagram, {('A', 'B'): 'none',
                                     ('B', 'C'): 'forward',
                                     ('C', 'D'): 'back',
                                     ('D', 'E'): 'both'})

    def test_edge_attribute(self):
        diagram = self.build('edge_attribute.diag')
        self.assertEdgeDir(diagram, {('A', 'B'): 'forward',
                                     ('B', 'C'): 'forward',
                                     ('C', 'D'): 'forward',
                                     ('D', 'E'): 'none',
                                     ('E', 'F'): 'both',
                                     ('F', 'G'): 'forward'})
        self.assertEdgeColor(diagram, {('A', 'B'): (255, 0, 0),  # red
                                       ('B', 'C'): (255, 0, 0),  # red
                                       ('C', 'D'): (255, 0, 0),  # red
                                       ('D', 'E'): (0, 0, 0),
                                       ('E', 'F'): (255, 0, 0),  # red
                                       ('F', 'G'): (0, 0, 0)})
        self.assertEdgeThick(diagram, {('A', 'B'): None,
                                       ('B', 'C'): None,
                                       ('C', 'D'): None,
                                       ('D', 'E'): None,
                                       ('E', 'F'): None,
                                       ('F', 'G'): 3})

    def test_folded_edge_diagram(self):
        diagram = self.build('folded_edge.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'D': (0, 1),
                                    'E': (0, 2), 'F': (1, 1),
                                    'Z': (0, 3)})

    def test_skipped_edge_right_diagram(self):
        diagram = self.build('skipped_edge_right.diag')
        self.assertEdgeSkipped(diagram, {('A', 'B'): False,
                                         ('A', 'C'): True,
                                         ('B', 'C'): False})

    def test_skipped_edge_rightdown_diagram(self):
        diagram = self.build('skipped_edge_rightdown.diag')
        self.assertEdgeSkipped(diagram, {('A', 'B'): False,
                                         ('A', 'C'): True,
                                         ('A', 'D'): False,
                                         ('B', 'C'): False,
                                         ('B', 'D'): False})

    def test_skipped_edge_up_diagram(self):
        diagram = self.build('skipped_edge_up.diag')
        self.assertEdgeSkipped(diagram, {('C', 'A'): True})

    def test_skipped_edge_down_diagram(self):
        diagram = self.build('skipped_edge_down.diag')
        self.assertEdgeSkipped(diagram, {('A', 'C'): True})

    def test_skipped_edge_leftdown_diagram(self):
        diagram = self.build('skipped_edge_leftdown.diag')
        self.assertEdgeSkipped(diagram, {('A', 'B'): False,
                                         ('B', 'C'): False,
                                         ('B', 'D'): False,
                                         ('C', 'G'): True,
                                         ('F', 'G'): False})

    @capture_stderr
    def test_skipped_edge_flowchart_rightdown_diagram(self):
        diagram = self.build('skipped_edge_flowchart_rightdown.diag')
        self.assertEdgeSkipped(diagram, {('A', 'B'): False,
                                         ('A', 'C'): False,
                                         ('A', 'D'): True,
                                         ('C', 'D'): False})

    @capture_stderr
    def test_skipped_edge_flowchart_rightdown2_diagram(self):
        diagram = self.build('skipped_edge_flowchart_rightdown2.diag')
        self.assertEdgeSkipped(diagram, {('B', 'C'): False,
                                         ('A', 'C'): True})

    def test_skipped_edge_portrait_right_diagram(self):
        diagram = self.build('skipped_edge_portrait_right.diag')
        self.assertEdgeSkipped(diagram, {('A', 'C'): True})

    def test_skipped_edge_portrait_rightdown_diagram(self):
        diagram = self.build('skipped_edge_portrait_rightdown.diag')
        self.assertEdgeSkipped(diagram, {('A', 'B'): False,
                                         ('A', 'C'): False,
                                         ('A', 'E'): True,
                                         ('B', 'D'): False,
                                         ('C', 'E'): False})

    def test_skipped_edge_portrait_leftdown_diagram(self):
        diagram = self.build('skipped_edge_portrait_leftdown.diag')
        self.assertEdgeSkipped(diagram, {('A', 'B'): False,
                                         ('B', 'C'): False,
                                         ('D', 'C'): True,
                                         ('D', 'E'): False})

    def test_skipped_edge_portrait_down_diagram(self):
        diagram = self.build('skipped_edge_portrait_down.diag')
        self.assertEdgeSkipped(diagram, {('A', 'B'): False,
                                         ('A', 'C'): True,
                                         ('B', 'C'): False})

    @capture_stderr
    def test_skipped_edge_portrait_flowchart_rightdown_diagram(self):
        diagram = self.build('skipped_edge_portrait_flowchart_rightdown.diag')
        self.assertEdgeSkipped(diagram, {('A', 'B'): False,
                                         ('A', 'C'): False,
                                         ('A', 'D'): True,
                                         ('C', 'D'): False})

    @capture_stderr
    def test_skipped_edge_portrait_flowchart_rightdown2_diagram(self):
        diagram = self.build('skipped_edge_portrait_flowchart_rightdown2.diag')
        self.assertEdgeSkipped(diagram, {('B', 'C'): False,
                                         ('A', 'C'): True})



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import re
import copy
from blockdiag.utils import images, unquote, urlutil, uuid, XY
from blockdiag.utils.compat import u
from blockdiag.utils.logging import warning
from blockdiag import noderenderer, plugins


class Base(object):
    basecolor = (255, 255, 255)
    textcolor = (0, 0, 0)
    fontfamily = None
    fontsize = None
    style = None
    int_attrs = ['colwidth', 'colheight', 'fontsize']

    @classmethod
    def set_default_color(cls, color):
        cls.basecolor = images.color_to_rgb(color)

    @classmethod
    def set_default_style(cls, style):
        cls.style = style

    @classmethod
    def set_default_text_color(cls, color):
        cls.textcolor = images.color_to_rgb(color)

    @classmethod
    def set_default_fontfamily(cls, fontfamily):
        cls.fontfamily = fontfamily

    @classmethod
    def set_default_fontsize(cls, fontsize):
        cls.fontsize = int(fontsize)

    @classmethod
    def clear(cls):
        cls.basecolor = (255, 255, 255)
        cls.textcolor = (0, 0, 0)
        cls.fontfamily = None
        cls.fontsize = None
        cls.style = None

    def duplicate(self):
        return copy.copy(self)

    def set_attribute(self, attr):
        name = attr.name
        value = unquote(attr.value)

        if name == 'class':
            if value in Diagram.classes:
                klass = Diagram.classes[value]
                self.set_attributes(klass.attrs)
            else:
                msg = "Unknown class: %s" % value
                raise AttributeError(msg)
        elif hasattr(self, "set_%s" % name):
            getattr(self, "set_%s" % name)(value)
        elif name in self.int_attrs:
            setattr(self, name, int(value))
        elif hasattr(self, name) and not callable(getattr(self, name)):
            setattr(self, name, value)
        else:
            class_name = self.__class__.__name__
            msg = "Unknown attribute: %s.%s" % (class_name, attr.name)
            raise AttributeError(msg)

    def set_attributes(self, attrs):
        for attr in attrs:
            self.set_attribute(attr)

    def set_style(self, value):
        if re.search('^(?:none|solid|dotted|dashed|\d+(,\d+)*)$', value, re.I):
            self.style = value.lower()
        else:
            class_name = self.__class__.__name__
            msg = "unknown %s style: %s" % (class_name, value)
            raise AttributeError(msg)


class Element(Base):
    namespace = {}
    int_attrs = Base.int_attrs + ['width', 'height']

    @classmethod
    def get(cls, elemid):
        if not elemid:
            elemid = uuid.generate()

        unquote_id = unquote(elemid)
        if unquote_id not in cls.namespace:
            obj = cls(elemid)
            cls.namespace[unquote_id] = obj

        return cls.namespace[unquote_id]

    @classmethod
    def clear(cls):
        super(Element, cls).clear()
        cls.namespace = {}
        cls.basecolor = (255, 255, 255)
        cls.textcolor = (0, 0, 0)

    def __init__(self, elemid):
        self.id = unquote(elemid)
        self.label = ''
        self.xy = XY(0, 0)
        self.group = None
        self.drawable = False
        self.order = 0
        self.color = self.basecolor
        self.width = None
        self.height = None
        self.colwidth = 1
        self.colheight = 1
        self.stacked = False

    def __repr__(self):
        _format = "<%s '%s' %s %dx%d at 0x%08x>"
        params = (self.__class__.__name__, self.id, str(self.xy),
                  self.colwidth, self.colheight, id(self))
        return _format % params

    def set_color(self, color):
        self.color = images.color_to_rgb(color)

    def set_textcolor(self, color):
        self.textcolor = images.color_to_rgb(color)


class DiagramNode(Element):
    shape = 'box'
    int_attrs = Element.int_attrs + ['rotate']
    linecolor = (0, 0, 0)
    label_orientation = 'horizontal'
    desctable = []
    attrname = {}

    @classmethod
    def set_default_shape(cls, shape):
        cls.shape = shape

    @classmethod
    def set_default_linecolor(cls, color):
        cls.linecolor = images.color_to_rgb(color)

    @classmethod
    def clear(cls):
        super(DiagramNode, cls).clear()
        cls.shape = 'box'
        cls.linecolor = (0, 0, 0)
        cls.label_orientation = 'horizontal'
        cls.desctable = ['numbered', 'label', 'description']
        cls.attrname = dict(numbered='No', label='Name',
                            description='Description')

    def __init__(self, elemid):
        super(DiagramNode, self).__init__(elemid)

        self.label = unquote(elemid) or ''
        self.numbered = None
        self.icon = None
        self.background = None
        self.description = None
        self.rotate = 0
        self.drawable = True
        self.href = None

        plugins.fire_node_event(self, 'created')

    def set_attribute(self, attr):
        if plugins.fire_node_event(self, 'attr_changing', attr):
            super(DiagramNode, self).set_attribute(attr)
            plugins.fire_node_event(self, 'attr_changed', attr)

    def set_linecolor(self, color):
        self.linecolor = images.color_to_rgb(color)

    def set_shape(self, value):
        if noderenderer.get(value):
            self.shape = value
        else:
            msg = "unknown node shape: %s" % value
            raise AttributeError(msg)

    def set_icon(self, value):
        if urlutil.isurl(value) or os.path.isfile(value):
            self.icon = value
        else:
            warning("icon image not found: %s", value)

    def set_background(self, value):
        if urlutil.isurl(value) or os.path.isfile(value):
            self.background = value
        else:
            warning("background image not found: %s", value)

    def set_stacked(self, _):
        self.stacked = True

    def set_label_orientation(self, value):
        value = value.lower()
        if value in ('horizontal', 'vertical'):
            self.label_orientation = value
        else:
            msg = "unknown label orientation: %s" % value
            raise AttributeError(msg)

    def to_desctable(self):
        attrs = []
        for name in self.desctable:
            value = getattr(self, name)
            if value is None:
                attrs.append(u(""))
            else:
                attrs.append(value)

        return attrs


class NodeGroup(Element):
    basecolor = (243, 152, 0)

    @classmethod
    def clear(cls):
        super(NodeGroup, cls).clear()
        cls.basecolor = (243, 152, 0)

    def __init__(self, elemid):
        super(NodeGroup, self).__init__(elemid)

        self.level = 0
        self.separated = False
        self.shape = 'box'
        self.thick = 3
        self.nodes = []
        self.edges = []
        self.icon = None
        self.orientation = 'landscape'
        self.href = None

    def duplicate(self):
        copied = super(NodeGroup, self).duplicate()
        copied.nodes = []
        copied.edges = []

        return copied

    def parent(self, level):
        if self.level < level:
            return None

        group = self
        while group.level != level:
            group = group.group

        return group

    def is_parent(self, other):
        parent = self.parent(other.level)
        return parent == other

    def traverse_nodes(self, preorder=False):
        for node in self.nodes:
            if isinstance(node, NodeGroup):
                if preorder:
                    yield node

                for subnode in node.traverse_nodes(preorder=preorder):
                    yield subnode

                if not preorder:
                    yield node
            else:
                yield node

    def traverse_edges(self, preorder=False):
        if preorder:
            for edge in self.edges:
                yield edge

        for group in self.traverse_groups(preorder):
            for edge in group.traverse_edges(preorder):
                yield edge

        if not preorder:
            for edge in self.edges:
                yield edge

    def traverse_groups(self, preorder=False):
        for node in self.traverse_nodes(preorder=preorder):
            if isinstance(node, NodeGroup):
                yield node

    def fixiate(self, fixiate_nodes=False):
        if self.separated:
            self.colwidth = 1
            self.colheight = 1

            return
        elif len(self.nodes) > 0:
            self.colwidth = max(x.xy.x + x.colwidth for x in self.nodes)
            self.colheight = max(x.xy.y + x.colheight for x in self.nodes)

        for node in self.nodes:
            if fixiate_nodes:
                node.xy = XY(self.xy.x + node.xy.x,
                             self.xy.y + node.xy.y)

            if isinstance(node, NodeGroup):
                node.fixiate(fixiate_nodes)

    def update_order(self):
        for i, node in enumerate(self.nodes):
            node.order = i

    def set_orientation(self, value):
        value = value.lower()
        if value in ('landscape', 'portrait'):
            self.orientation = value
        else:
            msg = "unknown diagram orientation: %s" % value
            raise AttributeError(msg)

    def set_shape(self, value):
        value = value.lower()
        if value in ('box', 'line'):
            self.shape = value
        else:
            msg = "unknown group shape: %s" % value
            raise AttributeError(msg)


class DiagramEdge(Base):
    basecolor = (0, 0, 0)
    namespace = {}

    @classmethod
    def get(cls, node1, node2):
        if node1 not in cls.namespace:
            cls.namespace[node1] = {}

        if node2 not in cls.namespace[node1]:
            obj = cls(node1, node2)
            cls.namespace[node1][node2] = obj

        return cls.namespace[node1][node2]

    @classmethod
    def find(cls, node1, node2=None):
        if node1 is None and node2 is None:
            return cls.find_all()
        elif isinstance(node1, NodeGroup):
            edges = cls.find(None, node2)
            edges = (e for e in edges if e.node1.group.is_parent(node1))
            return [e for e in edges if not e.node2.group.is_parent(node1)]
        elif isinstance(node2, NodeGroup):
            edges = cls.find(node1, None)
            edges = (e for e in edges if e.node2.group.is_parent(node2))
            return [e for e in edges if not e.node1.group.is_parent(node2)]
        elif node1 is None:
            return [e for e in cls.find_all() if e.node2 == node2]
        else:
            if node1 not in cls.namespace:
                return []

            if node2 is None:
                return cls.namespace[node1].values()

            if node2 not in cls.namespace[node1]:
                return []

        return cls.namespace[node1][node2]

    @classmethod
    def find_all(cls):
        for v1 in cls.namespace.values():
            for v2 in v1.values():
                yield v2

    @classmethod
    def find_by_level(cls, level):
        edges = []
        for e in cls.find_all():
            edge = e.duplicate()
            skips = 0

            if edge.node1.group.level < level:
                skips += 1
            else:
                while edge.node1.group.level != level:
                    edge.node1 = edge.node1.group

            if edge.node2.group.level < level:
                skips += 1
            else:
                while edge.node2.group.level != level:
                    edge.node2 = edge.node2.group

            if skips == 2:
                continue

            edges.append(edge)

        return edges

    @classmethod
    def clear(cls):
        super(DiagramEdge, cls).clear()
        cls.namespace = {}
        cls.basecolor = (0, 0, 0)

    def __init__(self, node1, node2):
        self.node1 = node1
        self.node2 = node2
        self.crosspoints = []
        self.skipped = 0

        self.label = None
        self.description = None
        self.dir = 'forward'
        self.color = self.basecolor
        self.hstyle = None
        self.folded = None
        self.thick = None

    def __repr__(self):
        _format = "<%s '%s' %s - '%s' %s at 0x%08x>"
        params = (self.__class__.__name__, self.node1.id, self.node1.xy,
                  self.node2.id, self.node2.xy, id(self))
        return _format % params

    def set_dir(self, value):
        value = value.lower()
        if value in ('back', 'both', 'none', 'forward'):
            self.dir = value
        elif value == '-<':
            self.dir = 'forward'
            self.hstyle = 'onemany'
        elif value == '>-':
            self.dir = 'back'
            self.hstyle = 'manyone'
        elif value == '>-<':
            self.dir = 'both'
            self.hstyle = 'manymany'
        elif value == '->':
            self.dir = 'forward'
        elif value == '<-':
            self.dir = 'back'
        elif value == '<->':
            self.dir = 'both'
        elif value == '--':
            self.dir = 'none'
        else:
            msg = "unknown edge dir: %s" % value
            raise AttributeError(msg)

    def set_color(self, color):
        self.color = images.color_to_rgb(color)

    def set_hstyle(self, value):
        value = value.lower()
        if value in ('generalization', 'composition', 'aggregation'):
            self.hstyle = value
        elif value == 'oneone':
            self.dir = 'none'
            self.hstyle = value
        elif value == 'onemany':
            self.dir = 'forward'
            self.hstyle = value
        elif value == 'manyone':
            self.dir = 'back'
            self.hstyle = value
        elif value == 'manymany':
            self.dir = 'both'
            self.hstyle = value
        else:
            msg = "unknown edge hstyle: %s" % value
            raise AttributeError(msg)

    def set_folded(self, _):
        self.folded = True

    def set_nofolded(self, _):
        self.folded = False

    def set_thick(self, _):
        self.thick = 3

    @property
    def direction(self):
        node1 = self.node1.xy
        node2 = self.node2.xy

        if node1.x > node2.x:
            if node1.y > node2.y:
                _dir = 'left-up'
            elif node1.y == node2.y:
                _dir = 'left'
            else:
                _dir = 'left-down'
        elif node1.x == node2.x:
            if node1.y > node2.y:
                _dir = 'up'
            elif node1.y == node2.y:
                _dir = 'same'
            else:
                _dir = 'down'
        else:
            if node1.y > node2.y:
                _dir = 'right-up'
            elif node1.y == node2.y:
                _dir = 'right'
            else:
                _dir = 'right-down'

        return _dir

    def to_desctable(self):
        label = "%s -> %s" % (self.node1.label, self.node2.label)
        return [label, self.description]


class Diagram(NodeGroup):
    _DiagramNode = DiagramNode
    _DiagramEdge = DiagramEdge
    _NodeGroup = NodeGroup

    classes = {}
    shadow_style = 'blur'
    linecolor = (0, 0, 0)
    int_attrs = (NodeGroup.int_attrs +
                 ['node_width', 'node_height', 'span_width', 'span_height'])

    @classmethod
    def clear(cls):
        super(Diagram, cls).clear()
        cls.shadow_style = 'blur'
        cls.linecolor = (0, 0, 0)
        cls.classes = {}

    def __init__(self):
        super(Diagram, self).__init__(None)

        self.config = None
        self.node_width = None
        self.node_height = None
        self.span_width = None
        self.span_height = None
        self.page_padding = None
        self.edge_layout = None

    def set_plugin(self, name, attrs, **kwargs):
        kwargs.update(dict([str(unquote(attr.name)), unquote(attr.value)]
                           for attr in attrs))
        plugins.load([name], diagram=self, **kwargs)

    def set_plugins(self, value, **kwargs):
        modules = [name.strip() for name in value.split(',')]
        plugins.load(modules, diagram=self, **kwargs)

    def set_default_shape(self, value):
        if noderenderer.get(value):
            DiagramNode.set_default_shape(value)
        else:
            msg = "unknown node shape: %s" % value
            raise AttributeError(msg)

    def set_default_label_orientation(self, value):
        value = value.lower()
        if value in ('horizontal', 'vertical'):
            DiagramNode.label_orientation = value
        else:
            msg = "unknown label orientation: %s" % value
            raise AttributeError(msg)

    def set_default_text_color(self, color):
        warning("default_text_color is obsoleted; use default_textcolor")
        self.set_default_textcolor(color)

    def set_default_textcolor(self, color):
        self.textcolor = images.color_to_rgb(color)
        self._DiagramNode.set_default_text_color(self.textcolor)
        self._NodeGroup.set_default_text_color(self.textcolor)
        self._DiagramEdge.set_default_text_color(self.textcolor)

    def set_default_node_color(self, color):
        color = images.color_to_rgb(color)
        self._DiagramNode.set_default_color(color)

    def set_default_node_style(self, value):
        if re.search('^(?:none|solid|dotted|dashed|\d+(,\d+)*)$', value, re.I):
            self._DiagramNode.set_default_style(value)
        else:
            msg = "unknown node style: %s" % value
            raise AttributeError(msg)

    def set_default_line_color(self, color):
        warning("default_line_color is obsoleted; use default_linecolor")
        self.set_default_linecolor(color)

    def set_default_linecolor(self, color):
        self.linecolor = images.color_to_rgb(color)
        self._DiagramNode.set_default_linecolor(self.linecolor)
        self._DiagramEdge.set_default_color(self.linecolor)

    def set_default_group_color(self, color):
        color = images.color_to_rgb(color)
        self._NodeGroup.set_default_color(color)

    def set_shape_namespace(self, value):
        noderenderer.set_default_namespace(value)

    def set_default_fontfamily(self, fontfamily):
        self._DiagramNode.set_default_fontfamily(fontfamily)
        self._NodeGroup.set_default_fontfamily(fontfamily)
        self._DiagramEdge.set_default_fontfamily(fontfamily)

    def set_default_fontsize(self, fontsize):
        self._DiagramNode.set_default_fontsize(fontsize)
        self._NodeGroup.set_default_fontsize(fontsize)
        self._DiagramEdge.set_default_fontsize(fontsize)

    def set_shadow_style(self, value):
        value = value.lower()
        if value in ('solid', 'blur', 'none'):
            self.shadow_style = value
        else:
            msg = "unknown shadow style: %s" % value
            raise AttributeError(msg)

    def set_edge_layout(self, value):
        value = value.lower()
        if value in ('normal', 'flowchart'):
            warning("edge_layout is very experimental feature!")

            self.edge_layout = value
        else:
            msg = "unknown edge layout: %s" % value
            raise AttributeError(msg)

    def set_fontsize(self, value):
        warning("fontsize is obsoleted; use default_fontsize")
        self.set_default_fontsize(int(value))



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import io
import sys
try:
    from configparser import SafeConfigParser
except ImportError:
    from ConfigParser import SafeConfigParser


class ConfigParser(SafeConfigParser):
    def __init__(self):
        if sys.version_info > (2, 6) and sys.version_info < (2, 7):
            # only for Python2.6
            # - dict_type argument is supported py2.6 or later
            # - SafeConfigParser of py2.7 uses OrderedDict as default
            from ordereddict import OrderedDict
            SafeConfigParser.__init__(self, dict_type=OrderedDict)
        else:
            SafeConfigParser.__init__(self)

    def read(self, path):
        fd = io.open(path, 'r', encoding='utf-8-sig')
        self.readfp(fd)
        fd.close()



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division
import re
import math
from itertools import tee
try:
    from future_builtins import zip
except ImportError:
    pass
from functools import partial, wraps
from PIL import Image, ImageDraw, ImageFont, ImageFilter
from blockdiag.imagedraw import base
from blockdiag.imagedraw.utils import memoize
from blockdiag.imagedraw.utils.ellipse import dots as ellipse_dots
from blockdiag.utils import images, Box, Size, XY
from blockdiag.utils.fontmap import parse_fontpath, FontMap
from blockdiag.utils.myitertools import istep, stepslice


# apply monkey patch to pillow
from blockdiag.imagedraw.utils import pillow
pillow.apply_patch()


def point_pairs(xylist):
    iterable = iter(xylist)
    for pt in iterable:
        if isinstance(pt, int):
            yield (pt, next(iterable))
        else:
            yield pt


def line_segments(xylist):
    p1, p2 = tee(point_pairs(xylist))
    next(p2)
    return zip(p1, p2)


def dashize_line(line, length):
    pt1, pt2 = line
    if pt1[0] == pt2[0]:  # holizonal
        if pt1[1] > pt2[1]:
            pt2, pt1 = line

        r = stepslice(range(pt1[1], pt2[1]), length)
        for y1, y2 in istep(n for n in r):
            yield [(pt1[0], y1), (pt1[0], y2)]

    elif pt1[1] == pt2[1]:  # vertical
        if pt1[0] > pt2[0]:
            pt2, pt1 = line

        r = stepslice(range(pt1[0], pt2[0]), length)
        for x1, x2 in istep(n for n in r):
            yield [(x1, pt1[1]), (x2, pt1[1])]
    else:  # diagonal
        if pt1[0] > pt2[0]:
            pt2, pt1 = line

        # DDA (Digital Differential Analyzer) Algorithm
        locus = []
        m = float(pt2[1] - pt1[1]) / float(pt2[0] - pt1[0])
        x = pt1[0]
        y = pt1[1]

        while x <= pt2[0]:
            locus.append((int(x), int(round(y))))
            x += 1
            y += m

        for p1, p2 in istep(stepslice(locus, length)):
            yield (p1, p2)


def style2cycle(style, thick):
    if thick is None:
        thick = 1

    if style == 'dotted':
        length = [2 * thick, 2 * thick]
    elif style == 'dashed':
        length = [4 * thick, 4 * thick]
    elif style == 'none':
        length = [0, 65535 * thick]
    elif re.search('^\d+(,\d+)*$', style or ""):
        length = [int(n) * thick for n in style.split(',')]
    else:
        length = None

    return length


def ttfont_for(font):
    if font.path:
        path, index = parse_fontpath(font.path)
        if index:
            ttfont = ImageFont.truetype(path, font.size, index=index)
        else:
            ttfont = ImageFont.truetype(path, font.size)
    else:
        ttfont = None

    return ttfont


class ImageDrawExBase(base.ImageDraw):
    def __init__(self, filename, **kwargs):
        self.filename = filename
        self.transparency = kwargs.get('transparency')
        self.bgcolor = kwargs.get('color', (256, 256, 256))
        self._image = None
        self.draw = None

        if kwargs.get('parent'):
            self.scale_ratio = kwargs.get('parent').scale_ratio
        else:
            self.scale_ratio = kwargs.get('scale_ratio', 1)

        self.set_canvas_size(Size(1, 1))  # This line make textsize() workable

    def paste(self, image, pt, mask=None):
        self._image.paste(image, pt, mask)
        self.draw = ImageDraw.Draw(self._image)

    def set_canvas_size(self, size):
        if self.transparency:
            mode = 'RGBA'
        else:
            mode = 'RGB'

        self._image = Image.new(mode, size, self.bgcolor)

        # set transparency to background
        if self.transparency:
            alpha = Image.new('L', size, 1)
            self._image.putalpha(alpha)

        self.draw = ImageDraw.Draw(self._image)

    def resizeCanvas(self, size):
        self._image = self._image.resize(size, Image.ANTIALIAS)
        self.draw = ImageDraw.Draw(self._image)

    def arc(self, box, start, end, **kwargs):
        style = kwargs.get('style')
        if 'style' in kwargs:
            del kwargs['style']
        if 'thick' in kwargs:
            del kwargs['thick']

        if style:
            while start > end:
                end += 360

            cycle = style2cycle(style, kwargs.get('width'))
            for pt in ellipse_dots(box, cycle, start, end):
                self.draw.line([pt, pt], fill=kwargs['fill'])
        else:
            self.draw.arc(box.to_integer_point(), start, end, **kwargs)

    def ellipse(self, box, **kwargs):
        if 'filter' in kwargs:
            del kwargs['filter']

        style = kwargs.get('style')
        if 'style' in kwargs:
            del kwargs['style']

        if style:
            if kwargs.get('fill') != 'none':
                kwargs2 = dict(kwargs)
                if 'outline' in kwargs2:
                    del kwargs2['outline']
                self.draw.ellipse(box, **kwargs2)

            if 'outline' in kwargs:
                kwargs['fill'] = kwargs['outline']
                del kwargs['outline']

            cycle = style2cycle(style, kwargs.get('width'))
            for pt in ellipse_dots(box, cycle):
                self.draw.line([pt, pt], fill=kwargs['fill'])
        else:
            if kwargs.get('fill') == 'none':
                del kwargs['fill']

            self.draw.ellipse(box.to_integer_point(), **kwargs)

    def line(self, xy, **kwargs):
        if 'jump' in kwargs:
            del kwargs['jump']
        if 'thick' in kwargs:
            if kwargs['thick'] is not None:
                kwargs['width'] = kwargs['thick']
            del kwargs['thick']

        style = kwargs.get('style')
        if kwargs.get('fill') == 'none':
            pass
        elif (style in ('dotted', 'dashed', 'none') or
              re.search('^\d+(,\d+)*$', style or "")):
            self.dashed_line(xy, **kwargs)
        else:
            if 'style' in kwargs:
                del kwargs['style']

            self.draw.line(xy, **kwargs)

    def dashed_line(self, xy, **kwargs):
        style = kwargs.get('style')
        del kwargs['style']

        cycle = style2cycle(style, kwargs.get('width'))
        for line in line_segments(xy):
            for subline in dashize_line(line, cycle):
                self.line(subline, **kwargs)

    def rectangle(self, box, **kwargs):
        thick = kwargs.get('thick', self.scale_ratio)
        fill = kwargs.get('fill')
        outline = kwargs.get('outline')
        style = kwargs.get('style')

        if thick == 1:
            d = 0
        else:
            d = int(math.ceil(thick / 2.0))

        if fill and fill != 'none':
            self.draw.rectangle(box, fill=fill)

        x1, y1, x2, y2 = box
        lines = (((x1, y1), (x2, y1)), ((x1, y2), (x2, y2)),  # horizonal
                 ((x1, y1 - d), (x1, y2 + d)),  # vettical (left)
                 ((x2, y1 - d), (x2, y2 + d)))  # vertical (right)

        for line in lines:
            self.line(line, fill=outline, width=thick, style=style)

    def polygon(self, xy, **kwargs):
        if 'filter' in kwargs:
            del kwargs['filter']

        if kwargs.get('fill') != 'none':
            kwargs2 = dict(kwargs)

            if 'style' in kwargs2:
                del kwargs2['style']
            if 'outline' in kwargs2:
                del kwargs2['outline']
            self.draw.polygon(xy, **kwargs2)

        if kwargs.get('outline'):
            kwargs['fill'] = kwargs['outline']
            del kwargs['outline']
            self.line(xy, **kwargs)

    @property
    def textfolder(self):
        textfolder = super(ImageDrawExBase, self).textfolder
        return partial(textfolder, scale=self.scale_ratio)

    @memoize
    def textlinesize(self, string, font):
        ttfont = ttfont_for(font)
        if ttfont is None:
            size = self.draw.textsize(string, font=None)

            font_ratio = font.size * 1.0 / FontMap.BASE_FONTSIZE
            size = Size(int(size[0] * font_ratio),
                        int(size[1] * font_ratio))
        else:
            size = Size(*ttfont.getsize(string))

        return size

    def text(self, xy, string, font, **kwargs):
        fill = kwargs.get('fill')
        ttfont = ttfont_for(font)

        if ttfont is None:
            if self.scale_ratio == 1 and font.size == FontMap.BASE_FONTSIZE:
                self.draw.text(xy, string, fill=fill)
            else:
                size = self.draw.textsize(string)
                image = Image.new('RGBA', size)
                draw = ImageDraw.Draw(image)
                draw.text((0, 0), string, fill=fill)
                del draw

                basesize = (size[0] * self.scale_ratio,
                            size[1] * self.scale_ratio)
                text_image = image.resize(basesize, Image.ANTIALIAS)
                self.paste(text_image, xy, text_image)
        else:
            size = ttfont.getsize(string)

            # Generate mask to support BDF(bitmap font)
            mask = Image.new('1', size)
            draw = ImageDraw.Draw(mask)
            draw.text((0, 0), string, fill='white', font=ttfont)

            # Rendering text
            filler = Image.new('RGB', size, fill)
            self.paste(filler, xy, mask)

    def textarea(self, box, string, font, **kwargs):
        if 'rotate' in kwargs and kwargs['rotate'] != 0:
            angle = 360 - int(kwargs['rotate']) % 360
            del kwargs['rotate']

            if angle in (90, 270):
                _box = Box(0, 0, box.height, box.width)
            else:
                _box = box

            text = ImageDrawEx(None, parent=self, transparency=True)
            text.set_canvas_size(_box.size)
            textbox = Box(0, 0, _box.width, _box.height)
            text.textarea(textbox, string, font, **kwargs)

            filler = Image.new('RGB', box.size, kwargs.get('fill'))
            self.paste(filler, box.topleft, text._image.rotate(angle))
            return

        lines = self.textfolder(box, string, font, **kwargs)

        if kwargs.get('outline'):
            outline = kwargs.get('outline')
            self.rectangle(lines.outlinebox, fill='white', outline=outline)

        rendered = False
        for string, xy in lines.lines:
            self.text(xy, string, font, **kwargs)
            rendered = True

        if not rendered and font.size > 0:
            _font = font.duplicate()
            _font.size = int(font.size * 0.8)
            self.textarea(box, string, _font, **kwargs)

    def image(self, box, url):
        try:
            image = images.open(url, mode='pillow')

            # resize image.
            w = min([box.width, image.size[0] * self.scale_ratio])
            h = min([box.height, image.size[1] * self.scale_ratio])
            image.thumbnail((w, h), Image.ANTIALIAS)

            # centering image.
            w, h = image.size
            if box.width > w:
                x = box[0] + (box.width - w) // 2
            else:
                x = box[0]

            if box.height > h:
                y = box[1] + (box.height - h) // 2
            else:
                y = box[1]

            if image.mode == 'P':
                # convert P to RGBA to masking transparent pixels
                image = image.convert('RGBA')

            if image.mode == 'RGBA':
                self.paste(image, (x, y), mask=image)
            else:
                self.paste(image, (x, y))
        except IOError:
            pass

    def save(self, filename, size, _format):
        if filename:
            self.filename = filename

        if size is None:
            x = int(self._image.size[0] / self.scale_ratio)
            y = int(self._image.size[1] / self.scale_ratio)
            size = (x, y)

        self._image.thumbnail(size, Image.ANTIALIAS)

        if self.filename:
            self._image.save(self.filename, _format)
            image = None
        else:
            from io import BytesIO
            tmp = BytesIO()
            self._image.save(tmp, _format)
            image = tmp.getvalue()

        return image


def blurred(fn):
    PADDING = 16

    def get_shape_box(*args):
        if fn.__name__ == 'polygon':
            xlist = [pt.x for pt in args[0]]
            ylist = [pt.y for pt in args[0]]
            return Box(min(xlist), min(ylist), max(xlist), max(ylist))
        else:
            return args[0]

    def get_abs_coordinate(box, *args):
        dx = box.x1 - PADDING
        dy = box.y1 - PADDING
        if fn.__name__ == 'polygon':
            return [pt.shift(-dx, -dy) for pt in args[0]]
        else:
            return box.shift(-dx, -dy)

    def create_shadow(self, size, *args, **kwargs):
        drawer = ImageDrawExBase(self.filename, transparency=True)
        drawer.set_canvas_size(size)
        getattr(drawer, fn.__name__)(*args, **kwargs)

        for _ in range(15):
            drawer._image = drawer._image.filter(ImageFilter.SMOOTH_MORE)

        return drawer._image

    @wraps(fn)
    def func(self, *args, **kwargs):
        args = list(args)

        if kwargs.get('filter') not in ('blur', 'transp-blur'):
            return fn(self, *args, **kwargs)
        else:
            box = get_shape_box(*args)
            args[0] = get_abs_coordinate(box, *args)

            size = Size(box.width + PADDING * 2, box.height + PADDING * 2)
            shadow = create_shadow(self, size, *args, **kwargs)
            xy = XY(box.x1 - PADDING, box.y1 - PADDING)
            self.paste(shadow, xy, shadow)

    return func


class ImageDrawEx(ImageDrawExBase):
    @blurred
    def ellipse(self, box, **kwargs):
        super(ImageDrawEx, self).ellipse(box, **kwargs)

    @blurred
    def rectangle(self, box, **kwargs):
        super(ImageDrawEx, self).rectangle(box, **kwargs)

    @blurred
    def polygon(self, xy, **kwargs):
        super(ImageDrawEx, self).polygon(xy, **kwargs)


def setup(self):
    from blockdiag.imagedraw import install_imagedrawer
    install_imagedrawer('png', ImageDrawEx)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re
from blockdiag.utils.compat import u, string_types
from io import StringIO


def _escape(s):
    if not isinstance(s, string_types):
        s = str(s)
    return s.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")


def _quote(s):
    return '"%s"' % _escape(s).replace('"', "&quot;")


class base(object):
    def __init__(self, *args, **kwargs):
        self.text = None
        self.elements = []
        self.attributes = {}
        for key, value in kwargs.items():
            self.add_attribute(key, value)

    def add_attribute(self, key, value):
        setter = 'set_%s' % key
        if hasattr(self, setter):
            getattr(self, setter)(value)
        else:
            key = re.sub('_', '-', key)
            self.attributes[key] = value

    def addElement(self, element):
        self.elements.append(element)

    def set_text(self, text):
        self.text = text

    def to_xml(self, io, level=0):
        clsname = self.__class__.__name__
        indent = '  ' * level

        io.write(u('%s<%s') % (indent, clsname))
        for key in sorted(self.attributes):
            value = self.attributes[key]
            if value is not None:
                io.write(u(' %s=%s') % (_escape(key), _quote(value)))

        if self.elements == []:
            if self.text is not None:
                io.write(u(">%s</%s>\n") % (_escape(self.text), clsname))
            else:
                io.write(u(" />\n"))
        elif self.elements:
            if self.text is not None:
                io.write(u(">%s\n") % (_escape(self.text),))
            else:
                io.write(u(">\n"))

            for e in self.elements:
                e.to_xml(io, level + 1)
            io.write(u('%s</%s>\n') % (indent, clsname))


class element(base):
    def __init__(self, x, y, width=None, height=None, *args, **kwargs):
        super(element, self).__init__(*args, **kwargs)
        self.attributes['x'] = x
        self.attributes['y'] = y
        if width is not None:
            self.attributes['width'] = width
        if height is not None:
            self.attributes['height'] = height


class svg(base):
    def __init__(self, x, y, width, height, **kwargs):
        if kwargs.get('noviewbox'):
            super(svg, self).__init__(width=(width - x), height=(height - y))
        else:
            viewbox = "%d %d %d %d" % (x, y, width, height)
            super(svg, self).__init__(viewBox=viewbox)

        self.nodoctype = kwargs.get('nodoctype', False)
        self.add_attribute('xmlns', 'http://www.w3.org/2000/svg')

    def to_xml(self):
        io = StringIO()

        if not self.nodoctype:
            url = "http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd"
            io.write(u("<?xml version='1.0' encoding='UTF-8'?>\n"))
            io.write(u('<!DOCTYPE svg PUBLIC ') +
                     u('"-//W3C//DTD SVG 1.0//EN" "%s">\n') % url)

        super(svg, self).to_xml(io)

        return io.getvalue()


class title(base):
    def __init__(self, _title):
        super(title, self).__init__(text=_title)


class desc(base):
    def __init__(self, _title):
        super(desc, self).__init__(text=_title)


class text(element):
    def __init__(self, x, y, _text, **kwargs):
        super(text, self).__init__(x, y, text=_text, **kwargs)


class rect(element):
    pass


class ellipse(base):
    def __init__(self, cx, cy, rx, ry, **kwargs):
        super(ellipse, self).__init__(cx=cx, cy=cy, rx=rx, ry=ry, **kwargs)


class image(element):
    def __init__(self, uri, x, y, width, height, **kwargs):
        super(image, self).__init__(x, y, width, height, **kwargs)
        self.add_attribute('xlink:href', uri)


class polygon(base):
    def __init__(self, points, **kwargs):
        xylist = " ".join('%d,%d' % pt for pt in points)
        super(polygon, self).__init__(points=xylist, **kwargs)


class path(base):
    def __init__(self, data, **kwargs):
        super(path, self).__init__(d=data, **kwargs)


class pathdata:
    def __init__(self, x=None, y=None):
        self.path = []
        if x is not None and y is not None:
            self.move(x, y)

    def closepath(self):
        self.path.append('z')

    def move(self, x, y):
        self.path.append('M %s %s' % (x, y))

    def relmove(self, x, y):
        self.path.append('m %s %s' % (x, y))

    def line(self, x, y):
        self.path.append('L %s %s' % (x, y))

    def relline(self, x, y):
        self.path.append('l %s %s' % (x, y))

    def hline(self, x):
        self.path.append('H%s' % (x,))

    def relhline(self, x):
        self.path.append('h%s' % (x,))

    def vline(self, y):
        self.path.append('V%s' % (y,))

    def relvline(self, y):
        self.path.append('v%s' % (y,))

    def bezier(self, x1, y1, x2, y2, x, y):
        self.path.append('C%s,%s %s,%s %s,%s' % (x1, y1, x2, y2, x, y))

    def relbezier(self, x1, y1, x2, y2, x, y):
        self.path.append('c%s,%s %s,%s %s,%s' % (x1, y1, x2, y2, x, y))

    def smbezier(self, x2, y2, x, y):
        self.path.append('S%s,%s %s,%s' % (x2, y2, x, y))

    def relsmbezier(self, x2, y2, x, y):
        self.path.append('s%s,%s %s,%s' % (x2, y2, x, y))

    def qbezier(self, x1, y1, x, y):
        self.path.append('Q%s,%s %s,%s' % (x1, y1, x, y))

    def qrelbezier(self, x1, y1, x, y):
        self.path.append('q%s,%s %s,%s' % (x1, y1, x, y))

    def smqbezier(self, x, y):
        self.path.append('T%s %s' % (x, y))

    def relsmqbezier(self, x, y):
        self.path.append('t%s %s' % (x, y))

    def ellarc(self, rx, ry, xrot, laf, sf, x, y):
        self.path.append('A%s,%s %s %s %s %s %s' %
                         (rx, ry, xrot, laf, sf, x, y))

    def relellarc(self, rx, ry, xrot, laf, sf, x, y):
        self.path.append('a%s,%s %s %s %s %s %s' %
                         (rx, ry, xrot, laf, sf, x, y))

    def __repr__(self):
        return ' '.join(self.path)


class defs(base):
    pass


class g(base):
    pass


class a(base):
    pass


class filter(element):
    def __init__(self, x, y, width, height, **kwargs):
        super(filter, self).__init__(x, y, width, height, **kwargs)


def svgclass(name):
    """ svg class generating function """
    return type(name, (base,), {})



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


def patch_FreeTypeFont_getsize():
    try:
        from PIL import PILLOW_VERSION
        from PIL.ImageFont import FreeTypeFont

        # Avoid offset problem in Pillow (>= 2.2.0, < 2.6.0)
        if "2.2.0" <= PILLOW_VERSION < "2.6.0":
            original_getsize = FreeTypeFont.getsize

            def getsize(self, string):
                size = original_getsize(self, string)
                offset = self.getoffset(string)

                return (size[0] + offset[0], size[1] + offset[1])

            FreeTypeFont.getsize = getsize
    except ImportError:
        pass


def apply_patch():
    patch_FreeTypeFont_getsize()



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from collections import defaultdict
from blockdiag.tests.utils import BuilderTestCase


class TestBuilderNode(BuilderTestCase):
    def test_single_node_diagram(self):
        diagram = self.build('single_node.diag')

        self.assertEqual(1, len(diagram.nodes))
        self.assertEqual(0, len(diagram.edges))
        self.assertEqual('A', diagram.nodes[0].label)
        self.assertEqual((0, 0), diagram.nodes[0].xy)

    def test_node_shape_diagram(self):
        expected = {'A': 'box', 'B': 'roundedbox', 'C': 'diamond',
                    'D': 'ellipse', 'E': 'note', 'F': 'cloud',
                    'G': 'mail', 'H': 'beginpoint', 'I': 'endpoint',
                    'J': 'minidiamond', 'K': 'flowchart.condition',
                    'L': 'flowchart.database', 'M': 'flowchart.input',
                    'N': 'flowchart.loopin', 'O': 'flowchart.loopout',
                    'P': 'actor', 'Q': 'flowchart.terminator', 'R': 'textbox',
                    'S': 'dots', 'T': 'none', 'U': 'square', 'V': 'circle',
                    'Z': 'box'}
        diagram = self.build('node_shape.diag')
        self.assertNodeShape(diagram, expected)

    def test_node_shape_namespace_diagram(self):
        diagram = self.build('node_shape_namespace.diag')
        self.assertNodeShape(diagram, {'A': 'flowchart.condition',
                                       'B': 'condition',
                                       'Z': 'box'})

    def test_node_has_multilined_label_diagram(self):
        diagram = self.build('node_has_multilined_label.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'Z': (0, 1)})
        self.assertNodeLabel(diagram, {'A': "foo\nbar", 'Z': 'Z'})

    def test_quoted_node_id_diagram(self):
        diagram = self.build('quoted_node_id.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), "'A'": (1, 0),
                                    'B': (2, 0), 'Z': (0, 1)})

    def test_node_id_includes_dot_diagram(self):
        diagram = self.build('node_id_includes_dot.diag')
        self.assertNodeXY(diagram, {'A.B': (0, 0), 'C.D': (1, 0),
                                    'Z': (0, 1)})

    def test_multiple_nodes_definition_diagram(self):
        diagram = self.build('multiple_nodes_definition.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (0, 1),
                                    'Z': (0, 2)})
        self.assertNodeColor(diagram, {'A': (255, 0, 0), 'B': (255, 0, 0),
                                       'Z': (255, 255, 255)})

    def test_multiple_node_relation_diagram(self):
        diagram = self.build('multiple_node_relation.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (1, 1), 'D': (2, 0),
                                    'Z': (0, 2)})

    def test_node_attribute(self):
        labels = {'A': 'B', 'B': 'double quoted', 'C': 'single quoted',
                  'D': '\'"double" quoted\'', 'E': '"\'single\' quoted"',
                  'F': 'F', 'G': 'G', 'H': 'H', 'I': 'I',
                  'J': 'Hello', 'K': 'K'}
        colors = {'A': (255, 0, 0), 'B': (255, 255, 255), 'C': (255, 0, 0),
                  'D': (255, 0, 0), 'E': (255, 0, 0), 'F': (255, 255, 255),
                  'G': (255, 255, 255), 'H': (255, 255, 255),
                  'I': (255, 255, 255), 'J': (255, 255, 255),
                  'K': (255, 255, 255)}
        textcolors = defaultdict(lambda: (0, 0, 0))
        textcolors['F'] = (255, 0, 0)
        linecolors = defaultdict(lambda: (0, 0, 0))
        linecolors['I'] = (255, 0, 0)
        numbered = defaultdict(lambda: None)
        numbered['E'] = '1'
        stacked = defaultdict(lambda: False)
        stacked['G'] = True
        fontsize = defaultdict(lambda: None)
        fontsize['H'] = 16
        orientations = defaultdict(lambda: 'horizontal')
        orientations['J'] = 'vertical'
        backgrounds = defaultdict(lambda: None)
        backgrounds['K'] = ('src/blockdiag/tests/diagrams/'
                            'debian-logo-256color-palettealpha.png')

        diagram = self.build('node_attribute.diag')
        self.assertNodeLabel(diagram, labels)
        self.assertNodeColor(diagram, colors)
        self.assertNodeTextColor(diagram, textcolors)
        self.assertNodeLineColor(diagram, linecolors)
        self.assertNodeNumbered(diagram, numbered)
        self.assertNodeStacked(diagram, stacked)
        self.assertNodeFontsize(diagram, fontsize)
        self.assertNodeLabel_Orientation(diagram, orientations)
        self.assertNodeBackground(diagram, backgrounds)

    def test_node_height_diagram(self):
        diagram = self.build('node_height.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'D': (2, 1),
                                    'E': (1, 1), 'Z': (0, 2)})

    def test_branched_diagram(self):
        diagram = self.build('branched.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'D': (1, 1),
                                    'E': (2, 1), 'Z': (0, 2)})

    def test_multiple_parent_node_diagram(self):
        diagram = self.build('multiple_parent_node.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (0, 2), 'D': (1, 2),
                                    'E': (0, 1), 'Z': (0, 3)})

    def test_twin_multiple_parent_node_diagram(self):
        diagram = self.build('twin_multiple_parent_node.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (0, 1), 'D': (1, 1),
                                    'E': (0, 2), 'Z': (0, 3)})

    def test_flowable_node_diagram(self):
        diagram = self.build('flowable_node.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'Z': (0, 1)})

    def test_plugin_autoclass_diagram(self):
        diagram = self.build('plugin_autoclass.diag')
        self.assertNodeXY(diagram, {'A_emphasis': (0, 0),
                                    'B_emphasis': (1, 0),
                                    'C': (1, 1)})
        self.assertNodeStyle(diagram, {'A_emphasis': 'dashed',
                                       'B_emphasis': 'dashed',
                                       'C': None})
        self.assertNodeColor(diagram, {'A_emphasis': (255, 0, 0),
                                       'B_emphasis': (255, 0, 0),
                                       'C': (255, 255, 255)})

    def test_plugin_attributes_diagram(self):
        diagram = self.build('plugin_attributes.diag')
        self.assertNodeTest_Attr1(diagram, {'A': "1", 'B': None})
        self.assertNodeTest_Attr2(diagram, {'A': "2", 'B': None})
        self.assertNodeTest_Attr3(diagram, {'A': "3", 'B': None})



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from blockdiag.noderenderer import install_renderer
from blockdiag.noderenderer.base import NodeShape
from blockdiag.utils import Box, XY


class Dots(NodeShape):
    def render_label(self, drawer, **kwargs):
        pass

    def render_shape(self, drawer, _, **kwargs):
        if kwargs.get('shadow'):
            return

        m = self.metrics
        center = m.cell(self.node).center
        dots = [center]
        if self.node.group.orientation == 'landscape':
            pt = XY(center.x, center.y - m.node_height / 2)
            dots.append(pt)

            pt = XY(center.x, center.y + m.node_height / 2)
            dots.append(pt)
        else:
            pt = XY(center.x - m.node_width / 3, center.y)
            dots.append(pt)

            pt = XY(center.x + m.node_width / 3, center.y)
            dots.append(pt)

        r = m.cellsize / 2
        for dot in dots:
            box = Box(dot.x - r, dot.y - r, dot.x + r, dot.y + r)
            drawer.ellipse(box, fill=self.node.linecolor,
                           outline=self.node.linecolor)


def setup(self):
    install_renderer('dots', Dots)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from functools import partial
from blockdiag.imagedraw import textfolder
from blockdiag.utils import Box


class ImageDraw(object):
    self_generative_methods = []
    nosideeffect_methods = ['set_canvas_size', 'textsize', 'textlinesize']
    supported_path = False
    baseline_text_rendering = False

    def set_canvas_size(self, size):
        pass

    def set_options(self, **kwargs):
        pass

    def line(self, xy, **kwargs):
        pass

    def rectangle(self, box, **kwargs):
        pass

    def polygon(self, xy, **kwargs):
        pass

    def arc(self, xy, start, end, **kwargs):
        pass

    def ellipse(self, xy, **kwargs):
        pass

    def textsize(self, string, font, maxwidth=None, **kwargs):
        if maxwidth is None:
            maxwidth = 65535

        box = Box(0, 0, maxwidth, 65535)
        textbox = self.textfolder(box, string, font, **kwargs)
        return textbox.outlinebox.size

    @property
    def textfolder(self):
        return partial(textfolder.get, self,
                       adjustBaseline=self.baseline_text_rendering)

    def textlinesize(self, string, font, **kwargs):
        pass

    def text(self, xy, string, font, **kwargs):
        pass

    def textarea(self, box, string, font, **kwargs):
        pass

    def image(self, box, url):
        pass

    def save(self, filename, size, _format):
        pass



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from hashlib import sha1
from docutils import nodes
import blockdiag.parser
import blockdiag.builder
import blockdiag.drawer


class blockdiag(nodes.General, nodes.Element):
    name = 'blockdiag'
    processor = blockdiag

    def to_diagram(self):
        try:
            tree = self.processor.parser.parse_string(self['code'])
        except:
            code = '%s { %s }' % (self.name, self['code'])
            tree = self.processor.parser.parse_string(code)
            self['code'] = code  # replace if succeeded

        return self.processor.builder.ScreenNodeBuilder.build(tree)

    def to_drawer(self, image_format, filename, fontmap, **kwargs):
        diagram = self.to_diagram()
        return self.processor.drawer.DiagramDraw(image_format, diagram,
                                                 filename, fontmap=fontmap,
                                                 **kwargs)

    def get_path(self, **options):
        options.update(self['options'])
        hashseed = (self['code'] + str(options)).encode('utf-8')
        hashed = sha1(hashseed).hexdigest()

        filename = "%s-%s.%s" % (self.name, hashed, options['format'].lower())
        outputdir = options.get('outputdir')
        if outputdir:
            filename = os.path.join(outputdir, filename)

        return filename



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import sys
import tempfile
from io import StringIO
from collections import namedtuple
from blockdiag.utils.compat import u
from blockdiag.utils.fontmap import FontInfo, FontMap
from blockdiag.tests.utils import capture_stderr

if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest


FontElement = namedtuple('FontElement', 'fontfamily fontsize')


class TestUtilsFontmap(unittest.TestCase):
    def setUp(self):
        fontpath1 = __file__
        fontpath2 = os.path.join(os.path.dirname(__file__), 'utils.py')
        self.fontpath = [fontpath1, fontpath2]

    def test_fontinfo_new(self):
        FontInfo("serif", None, 11)
        FontInfo("sansserif", None, 11)
        FontInfo("monospace", None, 11)
        FontInfo("cursive", None, 11)
        FontInfo("fantasy", None, 11)

        FontInfo("serif-bold", None, 11)
        FontInfo("sansserif-italic", None, 11)
        FontInfo("monospace-oblique", None, 11)
        FontInfo("my-cursive", None, 11)
        FontInfo("-fantasy", None, 11)

    def test_fontinfo_invalid_familyname1(self):
        with self.assertRaises(AttributeError):
            FontInfo("unknown", None, 11)

    def test_fontinfo_invalid_familyname2(self):
        with self.assertRaises(AttributeError):
            FontInfo("sansserif-", None, 11)

    def test_fontinfo_invalid_familyname3(self):
        with self.assertRaises(AttributeError):
            FontInfo("monospace-unkown", None, 11)

    def test_fontinfo_invalid_familyname4(self):
        with self.assertRaises(AttributeError):
            FontInfo("cursive-bold-bold", None, 11)

    def test_fontinfo_invalid_familyname5(self):
        with self.assertRaises(AttributeError):
            FontInfo("SERIF", None, 11)

    def test_fontinfo_invalid_fontsize1(self):
        with self.assertRaises(TypeError):
            FontInfo("serif", None, None)

    def test_fontinfo_invalid_fontsize2(self):
        with self.assertRaises(ValueError):
            FontInfo("serif", None, '')

    def test_fontinfo_parse(self):
        font = FontInfo("serif", None, 11)
        self.assertEqual('', font.name)
        self.assertEqual('serif', font.generic_family)
        self.assertEqual('normal', font.weight)
        self.assertEqual('normal', font.style)

        font = FontInfo("sansserif-bold", None, 11)
        self.assertEqual('', font.name)
        self.assertEqual('sans-serif', font.generic_family)
        self.assertEqual('bold', font.weight)
        self.assertEqual('normal', font.style)

        font = FontInfo("monospace-italic", None, 11)
        self.assertEqual('', font.name)
        self.assertEqual('monospace', font.generic_family)
        self.assertEqual('normal', font.weight)
        self.assertEqual('italic', font.style)

        font = FontInfo("my-cursive-oblique", None, 11)
        self.assertEqual('my', font.name)
        self.assertEqual('cursive', font.generic_family)
        self.assertEqual('normal', font.weight)
        self.assertEqual('oblique', font.style)

        font = FontInfo("my-fantasy-bold", None, 11)
        self.assertEqual('my', font.name)
        self.assertEqual('fantasy', font.generic_family)
        self.assertEqual('bold', font.weight)
        self.assertEqual('normal', font.style)

        font = FontInfo("serif-serif", None, 11)
        self.assertEqual('serif', font.name)
        self.assertEqual('serif', font.generic_family)
        self.assertEqual('normal', font.weight)
        self.assertEqual('normal', font.style)

    def test_fontinfo_familyname(self):
        font = FontInfo("serif", None, 11)
        self.assertEqual('serif-normal', font.familyname)

        font = FontInfo("sansserif-bold", None, 11)
        self.assertEqual('sansserif-bold', font.familyname)

        font = FontInfo("monospace-italic", None, 11)
        self.assertEqual('monospace-italic', font.familyname)

        font = FontInfo("my-cursive-oblique", None, 11)
        self.assertEqual('my-cursive-oblique', font.familyname)

        font = FontInfo("my-fantasy-bold", None, 11)
        self.assertEqual('my-fantasy-bold', font.familyname)

        font = FontInfo("serif-serif", None, 11)
        self.assertEqual('serif-serif-normal', font.familyname)

        font = FontInfo("-serif", None, 11)
        self.assertEqual('serif-normal', font.familyname)

    @capture_stderr
    def test_fontmap_empty_config(self):
        config = StringIO(u(""))
        fmap = FontMap(config)

        font1 = fmap.find()
        self.assertTrue(font1)
        self.assertEqual('sans-serif', font1.generic_family)
        self.assertEqual(None, font1.path)
        self.assertEqual(11, font1.size)

        element = FontElement('sansserif', 11)
        font2 = fmap.find(element)
        self.assertEqual(font1.familyname, font2.familyname)
        self.assertEqual(font1.path, font2.path)
        self.assertEqual(font1.size, font2.size)

        element = FontElement('sansserif-normal', 11)
        font3 = fmap.find(element)
        self.assertEqual(font1.familyname, font3.familyname)
        self.assertEqual(font1.path, font3.path)
        self.assertEqual(font1.size, font3.size)

        # non-registered familyname
        element = FontElement('my-sansserif-normal', 11)
        font4 = fmap.find(element)
        self.assertEqual(font1.familyname, font4.familyname)
        self.assertEqual(font1.path, font4.path)
        self.assertEqual(font1.size, font4.size)

    @capture_stderr
    def test_fontmap_none_config(self):
        fmap = FontMap()

        font1 = fmap.find()
        self.assertTrue(font1)
        self.assertEqual('sans-serif', font1.generic_family)
        self.assertEqual(None, font1.path)
        self.assertEqual(11, font1.size)

    def test_fontmap_normal_config(self):
        _config = u("[fontmap]\nsansserif: %s\nsansserif-bold: %s\n") % \
                  (self.fontpath[0], self.fontpath[1])
        config = StringIO(_config)
        fmap = FontMap(config)

        font1 = fmap.find()
        self.assertTrue(font1)
        self.assertEqual('sans-serif', font1.generic_family)
        self.assertEqual(self.fontpath[0], font1.path)
        self.assertEqual(11, font1.size)

        element = FontElement('sansserif', 11)
        font2 = fmap.find(element)
        self.assertEqual(font1.familyname, font2.familyname)
        self.assertEqual(font1.path, font2.path)
        self.assertEqual(font1.size, font2.size)

        element = FontElement('sansserif-normal', 11)
        font3 = fmap.find(element)
        self.assertEqual(font1.familyname, font3.familyname)
        self.assertEqual(font1.path, font3.path)
        self.assertEqual(font1.size, font3.size)

        element = FontElement('sansserif-bold', 11)
        font4 = fmap.find(element)
        self.assertEqual('sansserif-bold', font4.familyname)
        self.assertEqual(self.fontpath[1], font4.path)
        self.assertEqual(font1.size, font4.size)

        element = FontElement(None, None)
        font5 = fmap.find(element)
        self.assertEqual(font1.familyname, font5.familyname)
        self.assertEqual(font1.path, font5.path)
        self.assertEqual(font1.size, font5.size)

        element = object()
        font6 = fmap.find(element)
        self.assertEqual(font1.familyname, font6.familyname)
        self.assertEqual(font1.path, font6.path)
        self.assertEqual(font1.size, font6.size)

    def test_fontmap_duplicated_fontentry1(self):
        _config = u("[fontmap]\nsansserif: %s\nsansserif: %s\n") % \
                  (self.fontpath[0], self.fontpath[1])
        config = StringIO(_config)
        if sys.version_info[0] == 2:
            fmap = FontMap(config)

            font1 = fmap.find()
            self.assertEqual('sans-serif', font1.generic_family)
            self.assertEqual(self.fontpath[1], font1.path)
            self.assertEqual(11, font1.size)
        else:
            import configparser
            with self.assertRaises(configparser.DuplicateOptionError):
                FontMap(config)

    def test_fontmap_duplicated_fontentry2(self):
        # this testcase is only for python2.6 or later
        if sys.version_info > (2, 6):
            _config = u("[fontmap]\nsansserif: %s\nsansserif-normal: %s\n") % \
                      (self.fontpath[0], self.fontpath[1])
            config = StringIO(_config)
            fmap = FontMap(config)

            font1 = fmap.find()
            self.assertEqual('sans-serif', font1.generic_family)
            self.assertEqual(self.fontpath[1], font1.path)
            self.assertEqual(11, font1.size)

    def test_fontmap_with_capital_character(self):
        _config = u("[fontmap]\nCapitalCase-sansserif: %s\n") % \
                  self.fontpath[0]
        config = StringIO(_config)
        fmap = FontMap(config)

        element = FontElement('CapitalCase-sansserif', 11)
        font1 = fmap.find(element)
        self.assertEqual('sans-serif', font1.generic_family)
        self.assertEqual('capitalcase-sansserif-normal', font1.familyname)
        self.assertEqual(self.fontpath[0], font1.path)
        self.assertEqual(11, font1.size)

    @capture_stderr
    def test_fontmap_with_nodefault_fontentry(self):
        _config = u("[fontmap]\nserif: %s\n") % self.fontpath[0]
        config = StringIO(_config)
        fmap = FontMap(config)

        font1 = fmap.find()
        self.assertEqual('sans-serif', font1.generic_family)
        self.assertEqual(None, font1.path)
        self.assertEqual(11, font1.size)

        element = FontElement('serif', 11)
        font2 = fmap.find(element)
        self.assertEqual('serif', font2.generic_family)
        self.assertEqual(self.fontpath[0], font2.path)
        self.assertEqual(font1.size, font2.size)

        element = FontElement('fantasy', 20)
        font3 = fmap.find(element)
        self.assertEqual('sans-serif', font3.generic_family)
        self.assertEqual(None, font3.path)
        self.assertEqual(20, font3.size)

    @capture_stderr
    def test_fontmap_with_nonexistence_fontpath(self):
        _config = u("[fontmap]\nserif: unknown_file\n")
        config = StringIO(_config)
        fmap = FontMap(config)

        font1 = fmap.find()
        self.assertEqual('sans-serif', font1.generic_family)
        self.assertEqual(None, font1.path)
        self.assertEqual(11, font1.size)

    def test_fontmap_switch_defaultfamily(self):
        _config = u("[fontmap]\nserif-bold: %s\n") % self.fontpath[0]
        config = StringIO(_config)
        fmap = FontMap(config)

        font1 = fmap.find()
        self.assertEqual('sansserif-normal', font1.familyname)
        self.assertEqual(None, font1.path)
        self.assertEqual(11, font1.size)

        fmap.set_default_fontfamily('serif-bold')
        font2 = fmap.find()
        self.assertEqual('serif-bold', font2.familyname)
        self.assertEqual(self.fontpath[0], font2.path)
        self.assertEqual(11, font2.size)

        fmap.set_default_fontfamily('fantasy-italic')
        font3 = fmap.find()
        self.assertEqual('fantasy-italic', font3.familyname)
        self.assertEqual(None, font3.path)
        self.assertEqual(11, font3.size)

        fmap.fontsize = 20
        font4 = fmap.find()
        self.assertEqual('fantasy-italic', font4.familyname)
        self.assertEqual(None, font4.path)
        self.assertEqual(20, font4.size)

    def test_fontmap_using_fontalias(self):
        _config = (u("[fontmap]\nserif-bold: %s\n") +
                   u("[fontalias]\ntest = serif-bold\n")) % self.fontpath[0]
        config = StringIO(_config)
        fmap = FontMap(config)

        element = FontElement('test', 20)
        font1 = fmap.find(element)
        self.assertEqual('serif-bold', font1.familyname)
        self.assertEqual(self.fontpath[0], font1.path)
        self.assertEqual(20, font1.size)

    def test_fontmap_by_file(self):
        tmp = tempfile.mkstemp()

        _config = u("[fontmap]\nsansserif: %s\nsansserif-bold: %s\n") % \
                  (self.fontpath[0], self.fontpath[1])

        fp = os.fdopen(tmp[0], 'wt')
        fp.write(_config)
        fp.close()
        fmap = FontMap(tmp[1])

        font1 = fmap.find()
        self.assertTrue(font1)
        self.assertEqual('sans-serif', font1.generic_family)
        self.assertEqual(self.fontpath[0], font1.path)
        self.assertEqual(11, font1.size)

        os.unlink(tmp[1])

    def test_fontmap_including_bom_by_file(self):
        tmp = tempfile.mkstemp()

        _config = (u("[fontmap]\nsansserif: %s\n") +
                   u("sansserif-bold: %s\n")) % \
                  (self.fontpath[0], self.fontpath[1])

        try:
            fp = os.fdopen(tmp[0], 'wb')
            fp.write(_config.encode('utf-8-sig'))
            fp.close()
            fmap = FontMap(tmp[1])

            font1 = fmap.find()
            self.assertTrue(font1)
            self.assertEqual('sans-serif', font1.generic_family)
            self.assertEqual(self.fontpath[0], font1.path)
            self.assertEqual(11, font1.size)
        finally:
            os.unlink(tmp[1])



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re
import math
from reportlab.pdfgen import canvas
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont
from reportlab.lib.utils import ImageReader
from blockdiag.imagedraw import base
from blockdiag.imagedraw.utils import memoize
from blockdiag.utils import images, Box, Size
from blockdiag.utils.fontmap import parse_fontpath
from blockdiag.utils.compat import string_types


class PDFImageDraw(base.ImageDraw):
    baseline_text_rendering = True

    def __init__(self, filename, **kwargs):
        self.filename = filename
        self.canvas = None
        self.fonts = {}

        self.set_canvas_size(Size(1, 1))  # This line make textsize() workable

    def set_canvas_size(self, size):
        self.canvas = canvas.Canvas(self.filename, pagesize=size)
        self.size = size

    def set_font(self, font):
        if font.path is None:
            msg = "Could not detect fonts, use --font opiton\n"
            raise RuntimeError(msg)

        if font.path not in self.fonts:
            path, index = parse_fontpath(font.path)
            if index:
                ttfont = TTFont(font.path, path, subfontIndex=index)
            else:
                ttfont = TTFont(font.path, path)
            pdfmetrics.registerFont(ttfont)

            self.fonts[font.path] = ttfont

        self.canvas.setFont(font.path, font.size)

    def set_render_params(self, **kwargs):
        self.set_stroke_color(kwargs.get('outline'))
        self.set_fill_color(kwargs.get('fill', 'none'))
        self.set_style(kwargs.get('style'), kwargs.get('thick'))

        params = {}
        if kwargs.get('fill', 'none') == 'none':
            params['fill'] = 0
        else:
            params['fill'] = 1

        if kwargs.get('outline') is None:
            params['stroke'] = 0
        else:
            params['stroke'] = 1

        return params

    def set_style(self, style, thick):
        if thick is None:
            thick = 1

        if style == 'dotted':
            self.canvas.setDash([2 * thick, 2 * thick])
        elif style == 'dashed':
            self.canvas.setDash([4 * thick, 4 * thick])
        elif style == 'none':
            self.canvas.setDash([0, 65535 * thick])
        elif re.search('^\d+(,\d+)*$', style or ""):
            self.canvas.setDash([int(n) * thick for n in style.split(',')])
        else:
            self.canvas.setDash()

    def set_stroke_color(self, color="black"):
        if isinstance(color, string_types):
            self.canvas.setStrokeColor(color)
        elif color:
            rgb = (color[0] / 256.0, color[1] / 256.0, color[2] / 256.0)
            self.canvas.setStrokeColorRGB(*rgb)
        else:
            self.set_stroke_color()

    def set_fill_color(self, color="white"):
        if isinstance(color, string_types):
            if color != 'none':
                self.canvas.setFillColor(color)
        elif color:
            rgb = (color[0] / 256.0, color[1] / 256.0, color[2] / 256.0)
            self.canvas.setFillColorRGB(*rgb)
        else:
            self.set_fill_color()

    def path(self, pd, **kwargs):
        params = self.set_render_params(**kwargs)
        self.canvas.drawPath(pd, **params)

    def rectangle(self, box, **kwargs):
        x = box[0]
        y = self.size[1] - box[3]
        width = box[2] - box[0]
        height = box[3] - box[1]

        if 'thick' in kwargs and kwargs['thick'] is not None:
            self.canvas.setLineWidth(kwargs['thick'])

        params = self.set_render_params(**kwargs)
        self.canvas.rect(x, y, width, height, **params)

        if 'thick' in kwargs:
            self.canvas.setLineWidth(1)

    @memoize
    def textlinesize(self, string, font):
        self.set_font(font)
        width = self.canvas.stringWidth(string, font.path, font.size)
        return Size(int(math.ceil(width)), font.size)

    def text(self, xy, string, font, **kwargs):
        self.set_font(font)
        self.set_fill_color(kwargs.get('fill'))
        self.canvas.drawString(xy[0], self.size[1] - xy[1], string)

    def textarea(self, box, string, font, **kwargs):
        self.canvas.saveState()

        if 'rotate' in kwargs and kwargs['rotate'] != 0:
            angle = 360 - int(kwargs['rotate']) % 360
            self.canvas.rotate(angle)

            if angle == 90:
                box = Box(-box.y2, box.x1, -box.y1, box.x1 + box.width)
                box = box.shift(x=self.size.height, y=self.size.height)
            elif angle == 180:
                box = Box(-box.x2, -box.y2, -box.x1, -box.y2 + box.height)
                box = box.shift(y=self.size.height * 2)
            elif angle == 270:
                box = Box(box.y1, -box.x2, box.y2, -box.x1)
                box = box.shift(x=-self.size.height, y=self.size.height)

        self.set_font(font)
        lines = self.textfolder(box, string, font, **kwargs)

        if kwargs.get('outline'):
            outline = kwargs.get('outline')
            self.rectangle(lines.outlinebox, fill='white', outline=outline)

        rendered = False
        for string, xy in lines.lines:
            self.text(xy, string, font, **kwargs)
            rendered = True
        self.canvas.restoreState()

        if not rendered and font.size > 0:
            font.size = int(font.size * 0.8)
            self.textarea(box, string, font, **kwargs)

    def line(self, xy, **kwargs):
        self.set_stroke_color(kwargs.get('fill', 'none'))
        self.set_style(kwargs.get('style'), kwargs.get('thick'))

        if 'thick' in kwargs and kwargs['thick'] is not None:
            self.canvas.setLineWidth(kwargs['thick'])

        p1 = xy[0]
        y = self.size[1]
        for p2 in xy[1:]:
            self.canvas.line(p1.x, y - p1.y, p2.x, y - p2.y)
            p1 = p2

        if 'thick' in kwargs:
            self.canvas.setLineWidth(1)

    def arc(self, xy, start, end, **kwargs):
        start, end = 360 - end, 360 - start
        r = (360 + end - start) % 360

        self.set_render_params(**kwargs)
        y = self.size[1]
        self.canvas.arc(xy[0], y - xy[3], xy[2], y - xy[1], start, r)

    def ellipse(self, xy, **kwargs):
        params = self.set_render_params(**kwargs)
        y = self.size[1]
        self.canvas.ellipse(xy[0], y - xy[3], xy[2], y - xy[1], **params)

    def polygon(self, xy, **kwargs):
        pd = self.canvas.beginPath()
        y = self.size[1]
        pd.moveTo(xy[0][0], y - xy[0][1])
        for p in xy[1:]:
            pd.lineTo(p[0], y - p[1])

        params = self.set_render_params(**kwargs)
        self.canvas.drawPath(pd, **params)

    def image(self, box, url):
        try:
            image = images.open(url, mode='pillow')
            if image.mode not in ('RGBA', 'L', 'RGB', 'CYMYK'):
                # convert to format that reportlab can recognize
                image = image.convert('RGBA')

            y = self.size[1] - box[3]
            data = ImageReader(image)
            self.canvas.drawImage(data, box.x1, y, box.width, box.height,
                                  mask='auto', preserveAspectRatio=True)
        except IOError:
            pass

    def save(self, filename, size, _format):
        # Ignore size and format parameter; compatibility for ImageDrawEx.

        self.canvas.showPage()
        self.canvas.save()


def setup(self):
    from blockdiag.imagedraw import install_imagedrawer
    install_imagedrawer('pdf', PDFImageDraw)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division
from blockdiag.noderenderer import install_renderer
from blockdiag.noderenderer.base import NodeShape
from blockdiag.utils import images, Box, XY


class TextBox(NodeShape):
    def __init__(self, node, metrics=None):
        super(TextBox, self).__init__(node, metrics)

        if self.node.background:
            size = images.get_image_size(self.node.background)
            size = images.calc_image_size(size, self.textbox.size)

            pt = self.textbox.center
            self.textbox = Box(pt.x - size[0] // 2, pt.y - size[1] // 2,
                               pt.x + size[0] // 2, pt.y + size[1] // 2)

            self.connectors[0] = XY(pt.x, self.textbox[1])
            self.connectors[1] = XY(self.textbox[2], pt.y)
            self.connectors[2] = XY(pt.x, self.textbox[3])
            self.connectors[3] = XY(self.textbox[0], pt.y)

        if self.node.icon:
            self.connectors[3] = XY(self.iconbox[0], pt.y)

    def render_shape(self, drawer, _, **kwargs):
        if not kwargs.get('shadow') and self.node.background:
            drawer.image(self.textbox, self.node.background)


def setup(self):
    install_renderer('textbox', TextBox)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import print_function

import os
import re
import sys
import functools
from shutil import rmtree
from tempfile import mkdtemp, mkstemp
from blockdiag.builder import ScreenNodeBuilder
from blockdiag.parser import parse_file

if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest

try:
    # sys.stderr in py2.x allows mixture of str and unicode
    from cStringIO import StringIO
except ImportError:
    # sys.stderr in py3.x allows only str objects (disallow bytes objs)
    from io import StringIO


def supported_pil():
    try:
        from PIL import _imagingft
        _imagingft
        return True
    except:
        return False


def with_pil(fn):
    if not supported_pil():
        fn.__test__ = False

    return fn


def supported_pdf():
    try:
        import reportlab
        reportlab

        return True
    except:
        return False


def with_pdf(fn):
    if not supported_pdf():
        fn.__test__ = False

    return fn


def capture_stderr(func):
    def wrap(*args, **kwargs):
        try:
            stderr = sys.stderr
            sys.stderr = StringIO()

            func(*args, **kwargs)

            if re.search('(ERROR|Traceback)', sys.stderr.getvalue()):
                raise AssertionError('Caught error')
        finally:
            if sys.stderr.getvalue():
                print("---[ stderr ] ---")
                print(sys.stderr.getvalue())

            sys.stderr = stderr

    return functools.wraps(func)(wrap)


stderr_wrapper = capture_stderr   # FIXME: deprecated


class TemporaryDirectory(object):
    def __init__(self, suffix='', prefix='tmp', dir=None):
        self.name = mkdtemp(suffix, prefix, dir)

    def __del__(self):
        self.clean()

    def clean(self):
        if os.path.exists(self.name):
            rmtree(self.name)

    def mkstemp(self, suffix='', prefix='tmp', text=False):
        return mkstemp(suffix, prefix, self.name, text)


class BuilderTestCase(unittest.TestCase):
    def build(self, filename):
        basedir = os.path.dirname(__file__)
        pathname = os.path.join(basedir, 'diagrams', filename)
        return self._build(parse_file(pathname))

    def _build(self, tree):
        return ScreenNodeBuilder.build(tree)

    def __getattr__(self, name):
        if name.startswith('assertNode'):
            def asserter(diagram, attributes):
                attr_name = name.replace('assertNode', '').lower()
                print("[node.%s]" % attr_name)
                for node in (n for n in diagram.nodes if n.drawable):
                    print(node)
                    excepted = attributes[node.id]
                    self.assertEqual(excepted, getattr(node, attr_name))

            return asserter
        elif name.startswith('assertEdge'):
            def asserter(diagram, attributes):
                attr_name = name.replace('assertEdge', '').lower()
                print("[edge.%s]" % attr_name)
                for edge in diagram.edges:
                    print(edge)
                    expected = attributes[(edge.node1.id, edge.node2.id)]
                    self.assertEqual(expected, getattr(edge, attr_name))

            return asserter
        else:
            return getattr(super(BuilderTestCase, self), name)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division
from blockdiag.noderenderer import install_renderer
from blockdiag.noderenderer.base import NodeShape
from blockdiag.utils import Box, XY


class Square(NodeShape):
    def __init__(self, node, metrics=None):
        super(Square, self).__init__(node, metrics)

        r = min(metrics.node_width, metrics.node_height) // 2 + \
            metrics.cellsize // 2
        pt = metrics.cell(node).center
        self.connectors = [XY(pt.x, pt.y - r),  # top
                           XY(pt.x + r, pt.y),  # right
                           XY(pt.x, pt.y + r),  # bottom
                           XY(pt.x - r, pt.y)]  # left
        self.textbox = Box(pt.x - r, pt.y - r, pt.x + r, pt.y + r)

    def render_shape(self, drawer, _, **kwargs):
        fill = kwargs.get('fill')

        # draw outline
        if kwargs.get('shadow'):
            box = self.shift_shadow(self.textbox)
            if kwargs.get('style') == 'blur':
                drawer.rectangle(box, fill=fill, outline=fill,
                                 filter='transp-blur')
            else:
                drawer.rectangle(box, fill=fill, outline=fill)
        elif self.node.background:
            drawer.rectangle(self.textbox, fill=self.node.color,
                             outline=self.node.color)
            drawer.image(self.textbox, self.node.background)
            drawer.rectangle(self.textbox, fill="none",
                             outline=self.node.linecolor,
                             style=self.node.style)
        else:
            drawer.rectangle(self.textbox, fill=self.node.color,
                             outline=self.node.linecolor,
                             style=self.node.style)


def setup(self):
    install_renderer('square', Square)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import io
import sys
import tempfile
from blockdiag.tests.utils import with_pdf

import blockdiag
from blockdiag.command import BlockdiagOptions
from blockdiag.utils.bootstrap import detectfont
from blockdiag.utils.compat import u

if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest


class TestBootParams(unittest.TestCase):
    def setUp(self):
        self.parser = BlockdiagOptions(blockdiag)

    def test_type_option_svg(self):
        options = self.parser.parse(['-Tsvg', 'input.diag'])
        self.assertEqual(options.output, 'input.svg')

        options = self.parser.parse(['-TSVG', 'input.diag'])
        self.assertEqual(options.output, 'input.svg')

        options = self.parser.parse(['-TSvg', 'input.diag'])
        self.assertEqual(options.output, 'input.svg')

        options = self.parser.parse(['-TSvg', 'input.test.diag'])
        self.assertEqual(options.output, 'input.test.svg')

    def test_type_option_png(self):
        options = self.parser.parse(['-Tpng', 'input.diag'])
        self.assertEqual(options.output, 'input.png')

    @with_pdf
    def test_type_option_pdf(self):
        options = self.parser.parse(['-Tpdf', 'input.diag'])
        self.assertEqual(options.output, 'input.pdf')

    def test_invalid_type_option(self):
        with self.assertRaises(RuntimeError):
            self.parser.parse(['-Tsvgz', 'input.diag'])

    def test_separate_option_svg(self):
        self.parser.parse(['-Tsvg', '--separate', 'input.diag'])

    def test_separate_option_png(self):
        self.parser.parse(['-Tpng', '--separate', 'input.diag'])

    @with_pdf
    def test_separate_option_pdf(self):
        self.parser.parse(['-Tpdf', '--separate', 'input.diag'])

    def test_svg_nodoctype_option(self):
        self.parser.parse(['-Tsvg', '--nodoctype', 'input.diag'])

    def test_png_nodoctype_option(self):
        with self.assertRaises(RuntimeError):
            self.parser.parse(['-Tpng', '--nodoctype', 'input.diag'])

    def test_pdf_nodoctype_option(self):
        with self.assertRaises(RuntimeError):
            self.parser.parse(['-Tpdf', '--nodoctype', 'input.diag'])

    def test_svg_notransparency_option(self):
        with self.assertRaises(RuntimeError):
            self.parser.parse(['-Tsvg', '--no-transparency', 'input.diag'])

    def test_png_notransparency_option(self):
        self.parser.parse(['-Tpng', '--no-transparency', 'input.diag'])

    def test_pdf_notransparency_option(self):
        with self.assertRaises(RuntimeError):
            self.parser.parse(['-Tpdf', '--no-transparency', 'input.diag'])

    def test_config_option(self):
        try:
            tmp = tempfile.mkstemp()
            self.parser.parse(['-c', tmp[1], 'input.diag'])
        finally:
            os.close(tmp[0])
            os.unlink(tmp[1])

    def test_config_option_with_bom(self):
        try:
            tmp = tempfile.mkstemp()
            fp = io.open(tmp[0], 'wt', encoding='utf-8-sig')
            fp.write(u("[blockdiag]\n"))
            fp.close()

            self.parser.parse(['-c', tmp[1], 'input.diag'])
        finally:
            os.unlink(tmp[1])

    def test_invalid_config_option(self):
        with self.assertRaises(RuntimeError):
            self.parser.parse(['-c', '/unknown_config_file', 'input.diag'])

    def test_invalid_dir_config_option(self):
        try:
            tmp = tempfile.mkdtemp()

            with self.assertRaises(RuntimeError):
                self.parser.parse(['-c', tmp, 'input.diag'])
        finally:
            os.rmdir(tmp)

    def test_config_option_fontpath(self):
        try:
            tmp = tempfile.mkstemp()
            config = u("[blockdiag]\nfontpath = /path/to/font\n")
            io.open(tmp[0], 'wt', encoding='utf-8-sig').write(config)

            options = self.parser.parse(['-c', tmp[1], 'input.diag'])
            self.assertEqual(options.font, ['/path/to/font'])
        finally:
            os.unlink(tmp[1])

    def test_exist_font_config_option(self):
        try:
            fd, path = tempfile.mkstemp()
            os.close(fd)

            options = self.parser.parse(['-f', path, 'input.diag'])
            self.assertEqual(options.font, [path])
            fontpath = detectfont(options)
            self.assertEqual(fontpath, path)
        finally:
            os.unlink(path)

    def test_not_exist_font_config_option(self):
        with self.assertRaises(RuntimeError):
            args = ['-f', '/font_is_not_exist', 'input.diag']
            options = self.parser.parse(args)
            detectfont(options)

    def test_not_exist_font_config_option2(self):
        with self.assertRaises(RuntimeError):
            args = ['-f', '/font_is_not_exist',
                    '-f', '/font_is_not_exist2', 'input.diag']
            options = self.parser.parse(args)
            detectfont(options)

    def test_no_size_option(self):
        options = self.parser.parse(['input.diag'])
        self.assertEqual(None, options.size)

    def test_size_option(self):
        options = self.parser.parse(['--size', '480x360', 'input.diag'])
        self.assertEqual([480, 360], options.size)

    def test_invalid_size_option1(self):
        with self.assertRaises(RuntimeError):
            self.parser.parse(['--size', '480-360', 'input.diag'])

    def test_invalid_size_option2(self):
        with self.assertRaises(RuntimeError):
            self.parser.parse(['--size', '480', 'input.diag'])

    def test_invalid_size_option3(self):
        with self.assertRaises(RuntimeError):
            self.parser.parse(['--size', 'foobar', 'input.diag'])

    def test_not_exist_fontmap_config(self):
        with self.assertRaises(RuntimeError):
            args = ['--fontmap', '/fontmap_is_not_exist', 'input.diag']
            options = self.parser.parse(args)
            fontpath = detectfont(options)
            self.assertTrue(fontpath)

    def test_unknown_image_driver(self):
        from blockdiag.drawer import DiagramDraw
        from blockdiag.elements import Diagram

        with self.assertRaises(RuntimeError):
            DiagramDraw('unknown', Diagram())



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from blockdiag.noderenderer import install_renderer
from blockdiag.noderenderer.base import NodeShape
from blockdiag.utils import XY, Box
from blockdiag.imagedraw.simplesvg import pathdata


class RoundedBox(NodeShape):
    def render_shape(self, drawer, _, **kwargs):
        # draw background
        self.render_shape_background(drawer, **kwargs)

        # draw outline
        box = self.metrics.cell(self.node).box
        if not kwargs.get('shadow'):
            if self.node.background:
                drawer.image(box, self.node.background)

            self.render_shape_outline(drawer, **kwargs)

    def render_shape_outline(self, drawer, **kwargs):
        m = self.metrics.cell(self.node)
        r = self.metrics.cellsize
        box = m.box

        lines = [(XY(box[0] + r, box[1]), XY(box[2] - r, box[1])),
                 (XY(box[2], box[1] + r), XY(box[2], box[3] - r)),
                 (XY(box[0] + r, box[3]), XY(box[2] - r, box[3])),
                 (XY(box[0], box[1] + r), XY(box[0], box[3] - r))]
        for line in lines:
            drawer.line(line, fill=self.node.linecolor, style=self.node.style)

        r2 = r * 2
        arcs = [(Box(box[0], box[1], box[0] + r2, box[1] + r2), 180, 270),
                (Box(box[2] - r2, box[1], box[2], box[1] + r2), 270, 360),
                (Box(box[2] - r2, box[3] - r2, box[2], box[3]), 0, 90),
                (Box(box[0], box[3] - r2, box[0] + r2, box[3]), 90, 180)]
        for arc in arcs:
            drawer.arc(arc[0], arc[1], arc[2],
                       fill=self.node.linecolor, style=self.node.style)

    def render_shape_background(self, drawer, **kwargs):
        fill = kwargs.get('fill')

        m = self.metrics.cell(self.node)
        r = self.metrics.cellsize

        box = m.box
        ellipses = [Box(box[0], box[1], box[0] + r * 2, box[1] + r * 2),
                    Box(box[2] - r * 2, box[1], box[2], box[1] + r * 2),
                    Box(box[0], box[3] - r * 2, box[0] + r * 2, box[3]),
                    Box(box[2] - r * 2, box[3] - r * 2, box[2], box[3])]

        for e in ellipses:
            if kwargs.get('shadow'):
                e = self.shift_shadow(e)
                if kwargs.get('style') == 'blur':
                    drawer.ellipse(e, fill=fill, outline=fill,
                                   filter='transp-blur')
                else:
                    drawer.ellipse(e, fill=fill, outline=fill)
            else:
                drawer.ellipse(e, fill=self.node.color,
                               outline=self.node.color)

        rects = [Box(box[0] + r, box[1], box[2] - r, box[3]),
                 Box(box[0], box[1] + r, box[2], box[3] - r)]
        for rect in rects:
            if kwargs.get('shadow'):
                rect = self.shift_shadow(rect)
                if kwargs.get('style') == 'blur':
                    drawer.rectangle(rect, fill=fill, outline=fill,
                                     filter='transp-blur')
                else:
                    drawer.rectangle(rect, fill=fill, outline=fill)
            else:
                drawer.rectangle(rect, fill=self.node.color,
                                 outline=self.node.color)

    def render_vector_shape(self, drawer, _, **kwargs):
        fill = kwargs.get('fill')

        # create pathdata
        box = self.metrics.cell(self.node).box
        r = self.metrics.cellsize

        if kwargs.get('shadow'):
            box = self.shift_shadow(box)

        path = pathdata(box[0] + r, box[1])
        path.line(box[2] - r, box[1])
        path.ellarc(r, r, 0, 0, 1, box[2], box[1] + r)
        path.line(box[2], box[3] - r)
        path.ellarc(r, r, 0, 0, 1, box[2] - r, box[3])
        path.line(box[0] + r, box[3])
        path.ellarc(r, r, 0, 0, 1, box[0], box[3] - r)
        path.line(box[0], box[1] + r)
        path.ellarc(r, r, 0, 0, 1, box[0] + r, box[1])

        # draw outline
        if kwargs.get('shadow'):
            if kwargs.get('style') == 'blur':
                drawer.path(path, fill=fill, outline=fill,
                            filter='transp-blur')
            else:
                drawer.path(path, fill=fill, outline=fill)
        elif self.node.background:
            drawer.path(path, fill=self.node.color, outline=self.node.color)
            drawer.image(self.textbox, self.node.background)
            drawer.path(path, fill="none",
                        outline=self.node.linecolor, style=self.node.style)
        else:
            drawer.path(path, fill=self.node.color,
                        outline=self.node.linecolor, style=self.node.style)


def setup(self):
    install_renderer('roundedbox', RoundedBox)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import re
from PIL.Image import Image
from base64 import b64encode
from blockdiag.imagedraw import base as _base
from blockdiag.imagedraw.simplesvg import (
    svg, svgclass, filter, title, desc, defs, g, a, text,
    rect, polygon, ellipse, path, pathdata, image
)
from blockdiag.imagedraw.utils import memoize
from blockdiag.imagedraw.utils.ellipse import endpoints as ellipse_endpoints
from blockdiag.utils import images, Box, XY, is_Pillow_available

feGaussianBlur = svgclass('feGaussianBlur')


def rgb(color):
    if isinstance(color, tuple):
        color = 'rgb(%d,%d,%d)' % color

    return color


def style(name):
    if name == 'blur':
        value = "filter:url(#filter_blur)"
    elif name == 'transp-blur':
        value = "filter:url(#filter_blur);opacity:0.7;fill-opacity:1"
    else:
        value = None

    return value


def dasharray(pattern, thick):
    if thick is None:
        thick = 1

    if pattern == 'dotted':
        value = 2 * thick
    elif pattern == 'dashed':
        value = 4 * thick
    elif pattern == 'none':
        value = "%d %d" % (0, 65535 * thick)
    elif re.search('^\d+(,\d+)*$', pattern or ""):
        l = [int(n) * thick for n in pattern.split(",")]
        value = " ".join(str(n) for n in l)
    else:
        value = None

    return value


def drawing_params(kwargs):
    params = {}

    if 'style' in kwargs:
        params['stroke_dasharray'] = dasharray(kwargs.get('style'),
                                               kwargs.get('thick'))

    if 'filter' in kwargs:
        params['style'] = style(kwargs.get('filter'))

    return params


class SVGImageDrawElement(_base.ImageDraw):
    self_generative_methods = ['group', 'anchor']
    supported_path = True
    baseline_text_rendering = True

    def __init__(self, svg, parent=None):
        self.svg = svg

    def path(self, pd, **kwargs):
        fill = kwargs.get('fill')
        outline = kwargs.get('outline')

        p = path(pd, fill=rgb(fill), stroke=rgb(outline),
                 **drawing_params(kwargs))
        self.svg.addElement(p)

    def rectangle(self, box, **kwargs):
        thick = kwargs.get('thick')
        fill = kwargs.get('fill', 'none')
        outline = kwargs.get('outline')

        r = rect(box.x, box.y, box.width, box.height,
                 fill=rgb(fill), stroke=rgb(outline),
                 stroke_width=thick, **drawing_params(kwargs))
        self.svg.addElement(r)

    @memoize
    def textlinesize(self, string, font, **kwargs):
        if is_Pillow_available():
            if not hasattr(self, '_pil_drawer'):
                from blockdiag.imagedraw import png
                self._pil_drawer = png.ImageDrawEx(None)

            return self._pil_drawer.textlinesize(string, font)
        else:
            from blockdiag.imagedraw.utils import textsize
            return textsize(string, font)

    def text(self, point, string, font, **kwargs):
        fill = kwargs.get('fill')

        size = self.textlinesize(string, font)
        point = point.shift(size.width / 2)
        t = text(point.x, point.y, string, fill=rgb(fill),
                 font_family=font.generic_family, font_size=font.size,
                 font_weight=font.weight, font_style=font.style,
                 text_anchor='middle', textLength=size.width)
        self.svg.addElement(t)

    def textarea(self, box, string, font, **kwargs):
        if 'rotate' in kwargs and kwargs['rotate'] != 0:
            self.rotated_textarea(box, string, font, **kwargs)
        else:
            lines = self.textfolder(box, string, font, **kwargs)

            if kwargs.get('outline'):
                outline = kwargs.get('outline')
                self.rectangle(lines.outlinebox, fill='white', outline=outline)

            rendered = False
            for string, point in lines.lines:
                self.text(point, string, font, **kwargs)
                rendered = True

            if not rendered and font.size > 0:
                _font = font.duplicate()
                _font.size = int(font.size * 0.8)
                self.textarea(box, string, _font, **kwargs)

    def rotated_textarea(self, box, string, font, **kwargs):
        angle = int(kwargs['rotate']) % 360
        del kwargs['rotate']

        if angle in (90, 270):
            _box = Box(box[0], box[1],
                       box[0] + box.height, box[1] + box.width)
            if angle == 90:
                _box = _box.shift(x=box.width)
            elif angle == 270:
                _box = _box.shift(y=box.height)
        elif angle == 180:
            _box = Box(box[2], box[3],
                       box[2] + box.width, box[3] + box.height)
        else:
            _box = Box(box[0], box[1],
                       box[0] + box.width, box[1] + box.height)

        rotate = "rotate(%d,%d,%d)" % (angle, _box[0], _box[1])
        group = g(transform="%s" % rotate)
        self.svg.addElement(group)

        elem = SVGImageDrawElement(group, self)
        elem.textarea(_box, string, font, **kwargs)

    def line(self, points, **kwargs):
        fill = kwargs.get('fill')
        thick = kwargs.get('thick')

        pd = pathdata(points[0].x, points[0].y)
        for pt in points[1:]:
            pd.line(pt.x, pt.y)

        p = path(pd, fill="none", stroke=rgb(fill),
                 stroke_width=thick, **drawing_params(kwargs))
        self.svg.addElement(p)

    def arc(self, box, start, end, **kwargs):
        fill = kwargs.get('fill')

        w = box.width / 2
        h = box.height / 2

        if start > end:
            end += 360

        endpoints = ellipse_endpoints(1, w, h, start, end)
        pt1 = XY(box.x + w + round(endpoints[0].x, 0),
                 box.y + h + round(endpoints[0].y, 0))
        pt2 = XY(box.x + w + round(endpoints[1].x, 0),
                 box.y + h + round(endpoints[1].y, 0))

        if end - start > 180:
            largearc = 1
        else:
            largearc = 0

        pd = pathdata(pt1[0], pt1[1])
        pd.ellarc(w, h, 0, largearc, 1, pt2[0], pt2[1])
        p = path(pd, fill="none", stroke=rgb(fill),
                 **drawing_params(kwargs))
        self.svg.addElement(p)

    def ellipse(self, box, **kwargs):
        fill = kwargs.get('fill')
        outline = kwargs.get('outline')

        w = box.width / 2
        h = box.height / 2
        pt = box.center

        e = ellipse(pt.x, pt.y, w, h, fill=rgb(fill),
                    stroke=rgb(outline), **drawing_params(kwargs))
        self.svg.addElement(e)

    def polygon(self, points, **kwargs):
        fill = kwargs.get('fill')
        outline = kwargs.get('outline')

        pg = polygon(points, fill=rgb(fill), stroke=rgb(outline),
                     **drawing_params(kwargs))
        self.svg.addElement(pg)

    def image(self, box, url):
        if hasattr(url, 'read'):
            url = "data:;base64," + str(b64encode(url.read()))
        else:
            if isinstance(url, Image):
                ext = None
            else:
                ext = os.path.splitext(url)[1].lower()

            if ext not in ('.jpg', '.png', '.gif'):
                stream = None
                try:
                    stream = images.open(url, mode='png')
                    url = "data:;base64," + str(b64encode(stream.read()))
                except IOError:
                    pass
                finally:
                    if stream:
                        stream.close()

        im = image(url, box.x1, box.y1, box.width, box.height)
        self.svg.addElement(im)

    def anchor(self, url):
        a_node = a(url)
        a_node.add_attribute('xlink:href', url)
        self.svg.addElement(a_node)

        return SVGImageDrawElement(a_node, self)

    def group(self):
        group = g()
        self.svg.addElement(group)

        return SVGImageDrawElement(group, self)


class SVGImageDraw(SVGImageDrawElement):
    def __init__(self, filename, **kwargs):
        super(SVGImageDraw, self).__init__(None)

        self.filename = filename
        self.options = kwargs
        self.set_canvas_size((0, 0))

    def set_canvas_size(self, size):
        self.svg = svg(0, 0, size[0], size[1], **self.options)
        uri = 'http://www.inkscape.org/namespaces/inkscape'
        self.svg.add_attribute('xmlns:inkspace', uri)
        uri = 'http://www.w3.org/1999/xlink'
        self.svg.add_attribute('xmlns:xlink', uri)

        # inkspace's Gaussian filter
        if self.options.get('style') != 'blur':
            fgb = feGaussianBlur(id='feGaussianBlur3780', stdDeviation=4.2)
            fgb.add_attribute('inkspace:collect', 'always')
            f = filter(-0.07875, -0.252, 1.1575, 1.504, id='filter_blur')
            f.add_attribute('inkspace:collect', 'always')
            f.addElement(fgb)
            d = defs(id='defs_block')
            d.addElement(f)
            self.svg.addElement(d)

        self.svg.addElement(title('blockdiag'))
        self.svg.addElement(desc(self.options.get('code')))

    def save(self, filename, size, _format):
        # Ignore format parameter; compatibility for ImageDrawEx.

        if filename:
            self.filename = filename

        if size:
            self.svg.attributes['width'] = size[0]
            self.svg.attributes['height'] = size[1]

        image = self.svg.to_xml()

        if self.filename:
            open(self.filename, 'wb').write(image.encode('utf-8'))

        return image


def setup(self):
    from blockdiag.imagedraw import install_imagedrawer
    install_imagedrawer('svg', SVGImageDraw)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division
import re
import math


class XY(tuple):
    mapper = dict(x=0, y=1)

    def __new__(cls, x, y):
        return super(XY, cls).__new__(cls, (x, y))

    def __getattr__(self, name):
        try:
            return self[self.mapper[name]]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        raise TypeError("'XY' object does not support item assignment")

    def shift(self, x=0, y=0):
        return self.__class__(self.x + x, self.y + y)


class Size(tuple):
    def __new__(cls, width, height):
        return super(Size, cls).__new__(cls, (width, height))

    @property
    def width(self):
        return self[0]

    @property
    def height(self):
        return self[1]

    def resize(self, **kwargs):
        if 'width' in kwargs and 'height' in kwargs:
            size = Size(float(kwargs['width']), float(kwargs['height']))
        elif 'width' in kwargs:
            width = float(kwargs['width'])
            size = Size(width, self.height * width / self.width)
        elif 'height' in kwargs:
            height = float(kwargs['height'])
            size = Size(self.width * height / self.height, height)
        else:
            size = self

        if 'scale' in kwargs:
            scale = float(kwargs['scale']) / 100
            size = Size(size.width * scale, size.height * scale)

        return size

    def to_integer_point(self):
        return Size(int(self.width), int(self.height))


class Box(list):
    mapper = dict(x1=0, y1=1, x2=2, y2=3, x=0, y=1)

    def __init__(self, x1, y1, x2, y2):
        super(Box, self).__init__((x1, y1, x2, y2))

    def __getattr__(self, name):
        try:
            return self[self.mapper[name]]
        except KeyError:
            raise AttributeError(name)

    def __repr__(self):
        _format = "<%s (%s, %s) %dx%d at 0x%08x>"
        params = (self.__class__.__name__, self.x1, self.y1,
                  self.width, self.height, id(self))
        return _format % params

    def shift(self, x=0, y=0):
        return self.__class__(self.x1 + x, self.y1 + y,
                              self.x2 + x, self.y2 + y)

    def get_padding_for(self, size, **kwargs):
        valign = kwargs.get('valign', 'center')
        halign = kwargs.get('halign', 'center')
        padding = kwargs.get('padding', 0)

        if halign == 'left':
            dx = padding
        elif halign == 'right':
            dx = self.size.width - size.width - padding
        else:
            dx = int(math.ceil((self.size.width - size.width) / 2.0))

        if valign == 'top':
            dy = padding
        elif valign == 'bottom':
            dy = self.size.height - size.height - padding
        else:
            dy = int(math.ceil((self.size.height - size.height) / 2.0))

        return dx, dy

    @property
    def size(self):
        return Size(self.width, self.height)

    @property
    def width(self):
        return self.x2 - self.x1

    @property
    def height(self):
        return self.y2 - self.y1

    @property
    def topleft(self):
        return XY(self.x1, self.y1)

    @property
    def top(self):
        return XY(self.x1 + self.width // 2, self.y1)

    @property
    def topright(self):
        return XY(self.x2, self.y1)

    @property
    def bottomleft(self):
        return XY(self.x1, self.y2)

    @property
    def bottom(self):
        return XY(self.x1 + self.width // 2, self.y2)

    @property
    def bottomright(self):
        return XY(self.x2, self.y2)

    @property
    def left(self):
        return XY(self.x1, self.y1 + self.height // 2)

    @property
    def right(self):
        return XY(self.x2, self.y1 + self.height // 2)

    @property
    def center(self):
        return XY(self.x1 + self.width // 2, self.y1 + self.height // 2)

    def to_integer_point(self):
        return Box(*[int(i) for i in self])


def unquote(string):
    """ Remove quotas from string """
    if string:
        m = re.match('\A(?P<quote>"|\')((.|\s)*)(?P=quote)\Z', string, re.M)
        if m:
            return re.sub("\\\\" + m.group(1), m.group(1), m.group(2))
        else:
            return string
    else:
        return string


def is_Pillow_available():
    try:
        from PIL import _imagingft
        _imagingft
        return True
    except ImportError:
        return False
